import jsonpickle as jp
import re
from pyg_base._types import is_int, is_float, is_str, is_date, is_bool, is_pd, is_arr
from pyg_base._dates import dt, iso
from pyg_base._decorators import try_back
from pyg_base._logger import logger
from pyg_base._loop import loop
import pickle
import datetime
from functools import partial
from enum import Enum
import numpy as np


_obj = '_obj'
_data = 'data'
iso_quote = re.compile('^"[0-9]{4}-[0-9]{2}-[0-9]{2}T')

__all__ = ['encode', 'decode', 'pd2bson', 'bson2pd', 'bson2np']

@try_back
def decode_str(value):
    """
    A safer version of jp.decode

    :Parameters:
    ----------------
    value : str
        string to be decoded.

    :Returns:
    -------
    object
        value decoded or original value if failed.

    """
    res = jp.decode(value)
    if res is None and value!='null':
        logger.warning('could not decode value: %s'%value)
        return value
    else:
        return res



@loop(list, tuple)
def _decode(value, date = None):
    if is_str(value):
        if value.startswith('{'):
            value = decode_str(value)
            if not isinstance(value, str):
                value = _decode(value, date)
            return value
        elif value == 'null':
            return None
        if date in (None, False):
            return value
        elif date == 'iso' or date is True:
            if iso.search(value) is not None:
                return datetime.datetime.fromisoformat(value) 
            elif iso_quote.search(value) is not None:
                return datetime.datetime.fromisoformat(value.replace('"', ''))
            else:
                return value
        else:
            return value if date.search(value) is None else dt(value)
    elif isinstance(value, dict):
        res = type(value)(**{_decode(k, date) : _decode(v, date) for k, v in value.items()})
        return res.pop(_obj)(**res) if _obj in res.keys() else res
    else:
        return value
    
def decode(value, date = None):
    """
    decodes a string or an object dict 

    :Parameters:
    -------------
    value : str or dict
        usually a json
    date : None, bool or a regex expression, optional
        date format to be decoded
        
    :Returns:
    -------
    obj
        the json decoded.
    
    :Examples:
    --------------
    >>> from pyg import *
    >>> class temp(dict):
    >>>    pass
    
    >>> orig = temp(a = 1, b = dt(0))
    >>> encoded = encode(orig)
    >>> assert eq(decode(encoded), orig) # type matching too...
    
    """
    return _decode(value, date = date)

loads = partial(decode, date = True)

@loop(list, tuple)
def _encode(value, unchanged = None):
    if hasattr(value, '_encode') and not isinstance(value, type):
        res = value._encode
        if not isinstance(res, str):
            res = res()
        return res
    if is_bool(value):
        return True if value else False
    elif is_int(value):
        return int(value)
    elif is_float(value):
        return float(value)
    elif is_date(value):
        return value if isinstance(value, datetime.datetime) else dt(value) 
    elif isinstance(value, Enum):
        return _as_primitive(value.value)
    elif value is None or is_str(value):
        return value
    elif unchanged and isinstance(value, unchanged):
         return value
    elif isinstance(value, dict):
        res = {k : _encode(v, unchanged) for k, v in value.items()}
        if _obj not in res and type(value)!=dict:
            res[_obj] = _encode(type(value), unchanged)
        return res
    elif is_pd(value):
        return {_data : pd2bson(value), _obj : _bson2pd}
    elif is_arr(value):
        if value.dtype == np.dtype('O'):
            return {_data : pd2bson(value), _obj : _bson2pd}
        else:
            return {_data : value.tobytes(), 'shape' : value.shape, 'dtype' : encode(value.dtype), _obj : _bson2np}
    else:
        if hasattr(value, 'cache'):
            try:
                cache = value.cache
                del value.cache
                res = jp.encode(value)
                value.cache = cache
            except AttributeError:
                res = jp.encode(value)
        else:
            res = jp.encode(value)
        return res

@loop(list, tuple)
def _as_primitive(value):
    if is_bool(value):
        return True if value else False
    elif is_int(value):
        return int(value)
    elif is_float(value):
        return float(value)
    elif is_date(value):
        return dt(value)
    elif value is None or is_str(value):
        return value
    elif isinstance(value, Enum):
        return _as_primitive(value.value)
    else:
        return _encode(value)

def as_primitive(value):
    return _as_primitive(value)


def encode(value, unchanged = None):
    """
    
    encode/decode are performed prior to sending to mongodb or after retrieval from db. 
    The idea is to make object embedding in Mongo transparent to the user.
    
    - We use jsonpickle package to embed general objects. These are encoded as strings and can be decoded as long as the original library exists when decoding.
    - pandas.DataFrame are encoded to bytes using pickle while numpy arrays are encoded using the faster array.tobytes() with arrays' shape & type exposed and searchable.
    
    :Example:
    ----------
    >>> from pyg import *; import numpy as np
    >>> value = Dict(a=1,b=2)
    >>> assert encode(value) == {'a': 1, 'b': 2, '_obj': '{"py/type": "pyg_base._dict.Dict"}'}
    >>> assert decode({'a': 1, 'b': 2, '_obj': '{"py/type": "pyg_base._dict.Dict"}'}) == Dict(a = 1, b=2)
    >>> value = dictable(a=[1,2,3], b = 4)
    >>> assert encode(value) == {'a': [1, 2, 3], 'b': [4, 4, 4], '_obj': '{"py/type": "pyg_base._dictable.dictable"}'}
    >>> assert decode(encode(value)) == value
    >>> assert encode(np.array([1,2])) ==  {'data': bytes,
    >>>                                     'shape': (2,),
    >>>                                     'dtype': '{"py/reduce": [{"py/type": "numpy.dtype"}, {"py/tuple": ["i4", false, true]}, {"py/tuple": [3, "<", null, null, null, -1, -1, 0]}]}',
    >>>                                     '_obj': '{"py/function": "pyg_base._encode.bson2np"}'}
    
    :Example: functions and objects
    -------------------------------
    >>> from pyg import *; import numpy as np
    >>> assert encode(ewma) == '{"py/function": "pyg.timeseries._ewm.ewma"}'
    >>> assert encode(Calendar) == '{"py/type": "pyg_base._drange.Calendar"}'
    
    :Parameters:
    ----------------
    value : obj
        An object to be encoded 
        
    :Returns:
    -------
    A pre-json object

    """
    return _encode(value, unchanged)

def pd2bson(value):
    """
    converts a value (usually a pandas.DataFrame/Series) to bytes using pickle
    """
    return pickle.dumps(value)


def pd2pa(value):
    """serialize using arrow. Slightly slower than pickle"""
    import pyarrow as pa
    buf = pa.serialize(value).to_buffer()
    res = buf.to_pybytes()
    return res

def pa2pd(value):
    """serialize using arrow. Slightly slower than pickle"""
    import pyarrow as pa
    buf = pa.py_buffer(memoryview(value))
    df = pa.deserialize(buf)
    return df


def np2bson(value):
    """
    converts a numpy array to bytes using value.tobytes(). This is much faster than pickle but does not save shape/type info which we save separately.
    """
    return value.tobytes()

def bson2np(data, dtype, shape):
    """
    converts a byte with dtype and shape information into a numpy array.
    
    """
    res = np.frombuffer(data, dtype = dtype)
    return np.reshape(res, shape) if len(shape)!=1 else res

def bson2pd(data):
    """
    converts a pickled object back to an object. We insist that new object has .shape to ensure we did not unpickle gibberish.
    """
    try:
        res = pickle.loads(data)
        res.shape
        return res
    except Exception:
        return None        
    
_bson2pd = encode(bson2pd)
_bson2np = encode(bson2np)



from pyg_base._loop import loop
from pyg_base._types import is_str, is_num

_k = {0 : '', 3 : 'k', 6:'m', 9:'b', 12: 't', -2 : '%'}
_n = {v: 10**k for k,v in _k.items()}
_n.update({v.upper(): 10**k for k,v in _k.items()})

__all__ = ['as_float']

@loop(list, tuple, dict)
def _as_float(value):
    """
    converts a string to float

    :Parameters:
    ----------
    value : string
        number in string format.

    :Returns:
    -------
    float

    :Example:
    --------
    >>> from pyg import *
    >>> assert as_float('1.3k') == 1300
    >>> assert as_float('1.4m') == 1400000
    >>> assert as_float('100%') == 1
    >>> assert as_float('1,234') == 1234
    >>> assert as_float('-1,234k') == -1234000
    """
    if is_str(value):
        txt = value.replace(',','').replace(' ','')
        if not txt:
            return None
        if txt[-1] in _n:
            mult = _n[txt[-1]]
            txt = txt[:-1]
        else:
            mult = 1
        try:
            res = mult * float(txt)
            res = round(res, len(txt)+2)
            return res
        except ValueError:
            return value
    else:
        return value

def as_float(value):
    """
    converts a string to float allowing for commas, percentages etc.

    :Parameters:
    ----------
    value : string
        number in string format.

    :Returns:
    -------
    float

    :Example:
    --------
    >>> from pyg import *
    >>> assert as_float('1.3k') == 1300
    >>> assert as_float('1.4m') == 1400000
    >>> assert as_float('100%') == 1
    >>> assert as_float('1,234') == 1234
    >>> assert as_float('-1,234k') == -1234000
    """
    return _as_float(value)


import os
import json
CFG = os.environ.get('PYG_CFG', 'c:/etc/pyg.json')

def mkdir(path):
    """
    sames as os.mkdir but 
    1) allows for subdir i.e.:
    2) will not attempt to create if path exits, hence can run safely


    :Example:
    ---------
    >>> with pytest.raises(Exception):       
    >>>     os.mkdir('c:/temp/some/subdir/that/we/want')

    >>> print('but', mkdir('c:/temp/some/subdir/that/we/want'), 'now exists')
    
    """
    if not os.path.isdir(path):
        p = path.replace('\\', '/')
        paths = p.split('/')
        if '.' in paths[-1]:
            paths = paths[:-1]
        for i in range(2, len(paths)+1):
            d = '/'.join(paths[:i])
            if not os.path.isdir(d):
                os.mkdir(d)
    return path

def cfg_read():
    if os.path.isfile(CFG):
        with open(CFG, 'r') as f:
            cfg = json.load(f)
        return cfg
    else:
        return {}
cfg_read.__doc__ = 'reads the config file from %s' % CFG



def cfg_write(cfg):
    with open(mkdir(CFG), 'w') as f:
        json.dump(cfg, f)
cfg_write.__doc__ = 'writes the config file provided to %s' % CFG
    




from pyg_base._loop import loop
from pyg_base._types import is_str, is_float
from pyg_base._as_list import as_list, is_rng

alphabet = 'abcdefghijklmnopqrstuvwxyz'
ALPHABET = alphabet.upper()

__all__ = ['alphabet', 'ALPHABET', 'f12', 'replace', 'lower', 'upper', 'proper', 'strip', 'split', 'capitalize', 'common_prefix', 'deprefix']


@loop(list, dict, tuple)
def _f12(value):
    if is_float(value):
        return '%1.2f'%value
    else:
        return value

@loop(list, dict, tuple)
def _lower(text):
    return text.lower() if is_str(text) else text

@loop(list, dict, tuple)
def _upper(text):
    return text.upper() if is_str(text) else text


@loop(list, dict, tuple)
def _capitalize(text):
    return text.capitalize() if is_str(text) else text

_bbg = { k: ' ' + k.upper() for k in ['Comdty', 'Index', 'Curncy', 'Corp', 'Pfd', 'Equity', 'Govt', 'Mtge']}

@loop(list, dict, tuple)
def _bbgcase(text):
    if is_str(text):
        text = text.upper()
        for k, kupper  in _bbg.items():
            if text.endswith(kupper):
                return text.replace(kupper, ' ' + k)
            elif kupper + ' ' in text:
                return text.replace(kupper, ' ' + k)
    return text

@loop(list, dict, tuple)
def _proper(text):
    if is_str(text):
        return ' '.join([t.capitalize() for t in text.split(' ')])
    return text    
    
@loop(list, dict, tuple)
def _replace(text, old, new = None):
    if is_str(text):
        new = new or ''
        for arg in as_list(old):
            if arg in new:
                raise ValueError('cannot replace indefinitely "%s" with "%s"'%(arg, new))
            while arg in text:
                text = text.replace(arg, new)
    return text

@loop(list, dict, tuple)
def _strip(text, chars = None):
    if is_str(text):
        return text.strip(chars)
    return text


def replace(text, old, new = None):
    """
    A souped up version of text.replace(old, new)
    
    :Example: replace continues to replace until no-more is found
    --------------------------------------------------------------
    >>> assert replace('this    has lots  of   double    spaces', ' '*2, ' ') == 'this has lots of double spaces'
    >>> assert replace('this, sentence? has! too, many, punctuations!', list(',?!.')) == 'this sentence has too many punctuations'
    >>> assert replace(dict(a = 1, b = [' text within a list ', 'and within a dict']), ' ') == {'a': 1, 'b': ['textwithinalist', 'andwithinadict']}
    """
    return _replace(text, old = old, new = new)

def common_prefix(*values):
    """
    
    :Parameters:
    ----------------
    *values : list of iterables
        values for which we want to find common prefix

    :Returns:
    -------
    iterable
        the common prefix.
        
    :Example:
    --------------
    >>> assert common_prefix(['abra', 'abba', 'abacus']) == 'ab'
    >>> assert common_prefix('abra', 'abba', 'abacus') == 'ab'
    >>> assert common_prefix() is None
    >>> assert common_prefix([1,2,3,4], [1,2,3,5,8]) == [1,2,3]

    """
    values = as_list(values)
    if len(values) == 0:
        return None    
    def _all_match(i):
        i0 = values[0][i]
        for v in values:
            if v[i]!=i0:
                return False
        return True
    n = min([len(v) for v in values])
    i = 0
    while i<n and _all_match(i):
        i = i + 1
    return values[0][:i]

def deprefix(*values, sep = None):
    """
    remnoves the common prefix. If a sep is provided then remove only 'whole words' 
    

    :Parameters:
    ----------
    *values : TYPE
        list of values to be de-prefixed.
    sep : str, optional
        separations of wards in text. The default is None.

    :Returns:
    -------
    list
        de-prefixed values.

    :Example:
    ---------
    >>> assert deprefix('abra', 'abba') == ['ra', 'ba']
    >>> assert deprefix('abra', 'abba', sep ='_') == ['abra', 'abba']
    >>> assert deprefix('lhs_a', 'lhs_b', sep ='_') == ['a', 'b']
    
    """
    values = as_list(values)
    if sep is None:
        n = len(common_prefix(values))
        return [v[n:] for v in values]
    else:
        splitted = [v.split(sep) for v in values]
        res = deprefix(*splitted)
        return [sep.join(v) for v in res]
        

@loop(list, dict, tuple)
def _split(text, sep = ' ', dedup = False):
    if is_str(text):
        if is_rng(sep):
            if len(sep) == 0:
                sep = ' '
            elif len(sep) == 1:
                sep = sep[0]
            else:
                text = _replace(text, sep[1:], sep[0])
                sep = sep[0]
        res = text.split(sep)
        if dedup:
            res = [word for word in res if word]
        return res
    else:
        return text


def split(text, sep = ' ', dedup = False):
    """
    equivalent to txt.split(sep) but supporsts:
        - does not throw on non-string
        - removal of multiple seps
        - ensuring there is a unique single separator

    :Parameters:
    ----------
    text : str
        text to be stipped.
    sep : str, list of str, optional
        text used to strip. The default is ' '.
    dedup : bool, optional
        If True, will remove duplicated instances of seps. The default is False.

    :Returns:
    -------
    str
        splitted text
        
    :Example:
    ---------
    >>> text = '   The quick... brown .. fox... '    
    >>> assert split(text) == ['', '', '', 'The', 'quick...', 'brown', '..', 'fox...', '']
    >>> assert split(text, [' ', '.'], True) == ['The', 'quick', 'brown', 'fox']
    >>> text = dict(a = 'Can split this', b = '..and split this too')
    >>> assert split(text, [' ', '.'], True) == {'a': ['Can', 'split', 'this'], 'b': ['and', 'split', 'this', 'too']}
    """
    return _split(text, sep = sep, dedup = dedup)

def f12(value):
    return _f12(value)

def lower(value):
    """
    equivalent to txt.lower() but:
        - does not throw on non-string
        - supports lists/dicts
        
    :Example:
    ---------
    >>> assert lower(['The Brown Fox',1]) == ['the brown fox',1]
    >>> assert lower(dict(a = 'The Brown Fox', b = 3.0)) ==  {'a': 'the brown fox', 'b': 3.0}
    """
    return _lower(value)

def upper(value):
    """
    equivalent to txt.upper() but:
        - does not throw on non-string
        - supports lists/dicts
        
    :Example:
    ---------
    >>> assert upper(['The Brown Fox',1]) == ['THE BROWN FOX',1]
    >>> assert upper(dict(a = 'The Brown Fox', b = 3.0)) ==  {'a': 'THE BROWN FOX', 'b': 3.0}
    """
    return _upper(value)

def proper(value):
    """
    equivalent to Excel's PROPER(txt) but:
        - does not throw on non-string
        - supports lists/dicts
        
    :Example:
    ---------
    >>> assert proper(['THE BROWN FOX',1]) == ['The Brown Fox',1]
    >>> assert proper(dict(a = 'THE BROWN FOX', b = 3.0)) ==  {'a': 'The Brown Fox', 'b': 3.0}
    """
    return _proper(value)

def strip(value):
    """
    equivalent to txt.strip() but:
        - does not throw on non-string
        - supports lists/dicts
        
    :Example:
    ---------
    >>> assert strip([' whatever you say  ','  whatever you do..   ']) == ['whatever you say', 'whatever you do..']
    >>> assert strip(dict(a = ' whatever you say  ', b = 3.0)) ==  {'a': 'whatever you say', 'b': 3.0}
    """
    return _strip(value)

def bbgcase(value):
    return _bbgcase(value)

def capitalize(value):
    """
    equivalent to text.capitalize() but:
        - does not throw on non-string
        - supports lists/dicts
        
    :Example:
    ---------
    >>> assert capitalize('alan howard') == 'Alan howard' # use proper to get Alan Howard
    >>> assert capitalize(['alan howard', 'donald trump']) == ['Alan howard', 'Donald trump'] # use proper?
    """
    return _capitalize(value)
    


import os
import shutil
import re
from pyg_base._logger import logger
from pyg_npy import mkdir, path_name, path_dirname, path_join
import subprocess
import csv

__all__ = ['path_name', 'path_dirname', 'path_join', 'mkdir', 'read_csv']

    
def read_csv(path, errors = 'replace', **fmt):
    """
    A light-weight csv reader, no conversion is done, nor do we insist equal number of columns per row.
    - by default, encoding error (unicode characters) are replaced.
    - fmt parameters are parameters for the csv.reader object, see https://docs.python.org/3/library/csv.html
    """
    path = path_name(path)
    with open(path, 'r', errors = errors) as f:
        reader = csv.reader(f, **fmt)
        data = list(reader)
    return data


import inspect

__all__ = ['getargspec', 'getargs', 'getcallargs', 'call_with_callargs', 'argspec_defaults', 'argspec_required', 'argspec_update', 'kwargs2args', 'argspec_add']


def getargspec(function):
    """
    Extends inspect.getfullargspec to allow us to decorate functions with a signature.
    
    :Parameters:
    ----------------
    function : callable
        function for which we want to know argspec.

    :Returns:
    -------
    inspect.FullArgSpec

    """
    if hasattr(function, 'fullargspec'):
        return function.fullargspec
    else:
        return inspect.getfullargspec(function)


def argspec_add(fullargspec, **update):
    """
    adds new args with default values at the end of the existing args

    :Parameters:
    ----------
    fullargspec : FullArgSpec
        DESCRIPTION.
    **update : dict
        parameter names with their default values.

    :Returns:
    -------
    FullArgSpec

    :Example:
    ---------
    >>> f = lambda b : b
    >>> argspec = getargspec(f)
    >>> updated = argspec_add(argspec, axis = 0)
    >>> assert updated.args == ['b', 'axis'] and updated.defaults == (0,)

    >>> f = lambda b, axis : None ## axis already exists without a default
    >>> argspec = getargspec(f)
    >>> updated = argspec_add(argspec, axis = 0)
    >>> assert updated == argspec

    >>> f = lambda b, axis =1 : None ## axis already exists with a different default
    >>> argspec = getargspec(f)
    >>> updated = argspec_add(argspec, axis = 0)
    >>> assert updated == argspec


    """
    new_keys = [key for key in update if key not in fullargspec.args]
    if len(new_keys) == 0:
        return fullargspec
    new_defaults = tuple([update[key] for key in new_keys])
    args = fullargspec.args + new_keys
    defaults = (fullargspec.defaults or ()) + new_defaults 
    res = inspect.FullArgSpec(args = args, varargs = fullargspec.varargs, varkw = fullargspec.varkw, defaults = defaults, kwonlyargs = fullargspec.kwonlyargs, 
                               kwonlydefaults = fullargspec.kwonlydefaults, annotations = fullargspec.annotations)
    return res

def getargs(function, n = 0):
    """

    :Parameters:
    ----------------
    function : callable
        The function for  which we want the args
    n : int optional
        get the name of the args after allowing for first n args to be set by *args. 
        The default is 0.

    :Returns:
    -------
    None or a list of args

    """
    try:
        argspec = getargspec(function)
        if argspec.varargs or n == 0:
            return argspec.args
        else:
            return argspec.args[n:]
    except Exception:
        return []

def getcallargs(function, *args, **kwargs):
    """
    replicates inspect.getcallargs with support to functions within decorators
    
    :Example:
    --------------
    >>> from pyg import *; import inspect
    >>> function = lambda a, b, *myargs, **mykwargs: 1 
    >>> args = (1,2,3,4,5); kwargs = dict(c = 6, d = 7)
    >>> assert getcallargs(function, *args, **kwargs) == inspect.getcallargs(function, *args, **kwargs) == {'a': 1, 'b': 2, 'myargs': (3, 4, 5), 'mykwargs': {'c': 6, 'd': 7}} 
    
    >>> function = lambda a: a + 1
    >>> args = (); kwargs = dict(a=1)
    >>> assert getcallargs(function, *args, **kwargs) == inspect.getcallargs(function, *args, **kwargs) == dict(a = 1)

    >>> function = lambda a, b = 1: 1
    >>> args = (); kwargs = dict(a=1)
    >>> assert getcallargs(function, *args, **kwargs) == inspect.getcallargs(function, *args, **kwargs) == dict(a = 1, b = 1)
    >>> args = (); kwargs = dict(a=1, b = 2)
    >>> assert getcallargs(function, *args, **kwargs) == inspect.getcallargs(function, *args, **kwargs) == dict(a = 1, b = 2)
    >>> args = (1,); kwargs = {}
    >>> assert getcallargs(function, *args, **kwargs) == inspect.getcallargs(function, *args, **kwargs) == dict(a = 1, b = 1)
    >>> args = (1,2); kwargs = {}
    >>> assert getcallargs(function, *args, **kwargs) == inspect.getcallargs(function, *args, **kwargs) == dict(a = 1, b = 2)
    >>> args = (1,); kwargs = {'b' : 2}
    >>> assert getcallargs(function, *args, **kwargs) == inspect.getcallargs(function, *args, **kwargs) == dict(a = 1, b = 2)
    """
    spec = getargspec(function)
    arg_names = [] if spec.args is None else spec.args
    res = argspec_defaults(function)
    args2kwargs = dict(zip(arg_names, args))
    duplicates = set(args2kwargs) & set(kwargs)
    if len(duplicates):
        raise ValueError('got multiple values for arguments %s'%duplicates)
    res.update(args2kwargs)
    varargs = args[len(arg_names):]
    if spec.varargs is None:
        if len(varargs)>0:
            raise ValueError('The following varargs %s have no keywords'%varargs)
    else:
        res[spec.varargs] = varargs
    if spec.varkw is not None:
        varkw = {key: value for key, value in kwargs.items() if key not in arg_names}
        res[spec.varkw] = varkw
        res.update({key: value for key, value in kwargs.items() if key in arg_names})
    else:
        res.update(kwargs)
    return res

def call_with_callargs(function, callargs):
    """
    replicates inspect.getcallargs with support to functions within decorators
    
    :Example:
    --------------
    >>> function = lambda a, b, *args, **kwargs: 1+b+len(args)+10*len(kwargs)
    >>> args = (1,2,3,4,5); kwargs = dict(c = 6, d = 7)
    >>> assert function(*args, **kwargs) == 26
    >>> callargs = getcallargs(function, *args, **kwargs)
    >>> assert call_with_callargs(function, callargs) == 26
    """    
    if not callable(function):
        return function
    spec = getargspec(function)
    arg_names = [] if spec.args is None else spec.args
    c = dict(callargs)
    varargs = c.pop(spec.varargs) if spec.varargs else []
    varkw = c.pop(spec.varkw) if spec.varkw else {}
    defs = spec.defaults if spec.defaults else []
    params = dict(zip(arg_names[-len(defs):], defs))
    params.update(c)
    args = [params[arg] for arg in arg_names if arg in params] + list(varargs)
    return function(*args, **varkw)
        
        
    
    
def argspec_defaults(function):
    """
    :Returns: the function defaults as a dict rather than using the inspect structure
    
    :Example:
    --------------
    >>> f = lambda a, b = 1: a+b
    >>> assert argspec_defaults(f) == dict(b=1)

    >>> g = partial(f, b = 2)
    >>> assert argspec_defaults(g) == dict(b=2)

    
    :Parameters:
    ----------------
    function : callable

    :Returns:
    -------
    defaults as a dict.

    """
    argspec = getargspec(function)
    if argspec.defaults is None:
        res = {}
    else:
        res = dict(zip(argspec.args[-len(argspec.defaults):], argspec.defaults))
    if argspec.kwonlydefaults is not None:
        res.update(argspec.kwonlydefaults)
    return res

def argspec_required(function):
    """

    :Parameters:
    ----------------
    function : callable

    :Returns:
    -------
    list
        parameters that *must* be provided in order to run the function

    """
    
    
    argspec = getargspec(function)
    assert argspec.varargs is None, 'function cannot have unnamed *args and have a required args'
    return argspec.args if argspec.defaults is None else argspec.args[:-len(argspec.defaults)]  


def argspec_update(argspec, **kwargs):
    """
    generic function to create new FullArgSpec (python 3) or normal ArgSpec (python 2)

    :Parameters:
    ----------
    argspec : FullArgSpec
        The argspec of the dunction
    **kwargs : TYPE
        updates

    :Returns:
    -------
    FullArgSpec
    
    :Example:
    ---------
    >>> f = lambda a, b =1 : a + b
    >>> argspec = getargspec(f)    
    >>> assert argspec_update(argspec, args = ['a', 'b', 'c']) == inspect.FullArgSpec(**{'annotations': {},
                                                                                 'args': ['a', 'b', 'c'],
                                                                                 'defaults': (1,),
                                                                                 'kwonlyargs': [],
                                                                                 'kwonlydefaults': None,
                                                                                 'varargs': None,
                                                                                 'varkw': None})
    """
    tp = type(argspec)
    params = {key : getattr(argspec, key) for key in dir(tp)  if not key.startswith('_') and key not in ('count', 'index') }
    params.update(kwargs)
    return tp(**params)


def kwargs2args(function, args, kwargs):
    """
    converts a list of paramters that were provided as kwargs, into args

    :Example:
    -------
    >>> assert kwargs2args(lambda a, b: a+b, (), dict(a = 1, b=2)) == ([1,2], {})


    :Parameters:
    ----------------
    function : callable
    args : tuple
        parameters of function.
    kwargs : dict
        key-word parameters of function.

    :Returns:
    -------
    tuple
        a pair of a function args, kwargs.

    """
    if args:
        return args, kwargs
    else:
        kwargs = dict(kwargs)
        return [kwargs.pop(a) for a in getargspec(function).args if a in kwargs], kwargs





import logging

__all__ = ['get_logger', 'logger']

_loggers = dict()

def get_logger(name = 'pyg', level = 'info', fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s', file = False, console = True):
    """
    quick utility to simplify loggers creation and ensure we cache them and do not add to many handlers

    :Parameters:
    ----------
    name : str, optional
        name of logger. The default is 'pyg'.
    level : str, optional
        DEBUG/INFO/WARN etc. The default is 'info'.
    fmt : str, optional
        string formatting for messages. The default is '%(asctime)s - %(name)s - %(levelname)s - %(message)s'.
    file : bool/str, optional
        the name of the file to log to. The default is False = do not log to file.
    console : bool, optional
        log to console? The default is True.

    :Returns:
    -------
    logging.logger
    """
    if name in _loggers :
        return _loggers[name]
    logger = logging.getLogger(name)
    level = getattr(logging, level.upper()) if isinstance(level, str) else level
    logger.setLevel(level)
    formatter = logging.Formatter(fmt)
    if console:
        c = logging.StreamHandler()
        c.setLevel(level)
        c.setFormatter(formatter)
        logger.addHandler(c)
    if file:
        f = logging.FileHandler(file)
        f.setLevel(level)
        f.setFormatter(formatter)
        logger.addHandler(f)
    _loggers[name] = logger
    return logger

logger = get_logger()



from collections.abc import Hashable
from pyg_base._types import is_primitive
from pyg_base._decorators import wrapper, getargs

_cache = 'cache'
def _prehash(value):
    if isinstance(value, (tuple,list)):
        return tuple([_prehash(v) for v in value])
    elif isinstance(value, dict):
        return tuple(sorted([(k, _prehash(v)) for k, v in value.items()]))
    else:
        return value

CACHE = {}

def get_cache(*args):
    res = CACHE
    for arg in args:
        if arg not in res:
            res[arg] = {}
        res = res[arg]
    return res

def _hash(value):
    value = _prehash(value)
    if is_primitive(value):
        return value
    else:
        return hash(value)
    
class cache_func(wrapper):
    """
    >>> @cache
    >>> def f(a,b):
    >>>     return a+b

    >>> assert f(1,2) == 3
    >>> key = ((1, 2), ())
    >>> assert key in f.cache
    >>> assert f.cache[key] == 3

    """
    def _key(self, *args, **kwargs):
        return _prehash((args, kwargs))

    def wrapped(self, *args, **kwargs):
        key = self._key(*args, **kwargs)
        try:
            self.cache = getattr(self, _cache, {})
            if key not in self.cache:
                self.cache[key] = self.function(*args, **kwargs)
            return self.cache[key]
        except Exception:
            return self.function(*args, **kwargs)
        
    def clear_cache(self):
        self.cache = {}
        return self
    

# class cache_method(wrapper):
#     @property
#     def name(self):
#         return self.function.__name__
    
#     def _key(self, *args, **kwargs):
#         return _prehash((self.name, args[1:], kwargs))

#     def wrapped(self, *args, **kwargs):
#         key = self._key(*args, **kwargs)
#         me = args[0]
#         try:
#             me.cache = getattr(me, _cache, {})
#             if key not in me.cache:
#                 me.cache[key] = self.function(*args, **kwargs)
#             return me.cache[key]
#         except Exception:
#             return self.function(*args, **kwargs)
        
#     def clear_cache(self):
#         self.cache = {}
#         return self

def cache(function):
    """
    :Example:
    ---------
    >>> @cache
    >>> def f(a,b):
    >>>     return a+b

    >>> assert f(1,2) == 3
    >>> key = ((1, 2), ())
    >>> assert key in f.cache
    >>> assert f.cache[key] == 3

    """
    args = getargs(function)
    if args and args[0] in ('self', 'cls'):
        raise ValueError('cannot cache method')
    else:
        return cache_func(function)
    


    



from pyg_base._dict import dict_invert
from pyg_base._as_list import as_list
from pyg_base._cache import get_cache
from _collections_abc import dict_keys, dict_values


"""
This module is designed to remove dependency on networkx as it is slow to load and run
Since we only use topological_sort, our implementation is more specialized and hence faster (we don't test for cyclic nature of graph etc.)
"""

def get_DAG(dag = None):
    return dag if isinstance(dag, dict) else get_cache('DAG', dag)

def get_DAGs():
    return get_cache('DAG')

def get_GAD(dag = None):
    return dag if isinstance(dag, dict) else get_cache('GAD', dag)

def get_GADs():
    return get_cache('GAD')

def topological_sort(dag, node):    
    """
    This is a breadth-first topological sort, returning 
    The breadth-first is faster for even medium-sized graphs and also needs no inversion.

    The directed acyclic graph is stored as a dict with edges that look like:
        dag[parent] = {child : None for child in (parent, child) set of edges}

    :Parameters:
    ----------
    dag : dict, a directed acyclic graph
        DESCRIPTION.
    node : a key
        a node in the graph acting as the top of the subgraph

    :Returns:
    -------
    a dict of dicts: 
        node2gen: from nodes to their generations
        gen2node: from a generation to a set of its node

    :Example: constructing a DAG:
    ---------
    >>> dag = dict(a = dict(b = 0, c = 0), b = dict(c = 0, d = 0, e = 0), c = dict(e = 0), d = dict(e = 0))
    >>> node = 'a'
    >>> assert topological_sort(dag, node)['node2gen'] == {'a': 0, 'b': 1, 'c': 2, 'e': 3, 'd': 2}


    :Example: multiple nodes descendants:
    --------------------------------------
    >>> assert descendants(dag, ['c', 'd']) == ['c', 'd', 'e'] 
    
    NOTE: we cannot provide the list as a tuple. A tuple is immutable in itself and will be considered to be a key in the dag.
    

    :Example: we compare to networkx alternative:
    ---------
    >>> dag = dict(a = dict(b = 0, c = 0), b = dict(c = 0, d = 0, e = 0), c = dict(e = 0), d = dict(e = 0))
    >>> node = 'a'
    >>> ts = topological_sort(dag, node)
    >>> assert ts['node2gen'] == {'a': 0, 'b': 1, 'c': 2, 'e': 3, 'd': 2}

    >>> import networkx as nx
    >>> G = nx.DiGraph()
    >>> for a in dag:
    >>>     for b in dag[a]:
    >>>         G.add_edge(a,b)

    >>> def nx_topological_sort(G, node):
    >>>    nodes = {node} | nx.descendants(G, node)
    >>>    return list(nx.algorithms.dag.topological_sort(G.subgraph(nodes)))


    >>> def compare_topological_sort_and_nx(gen2node,y):
    >>>     if 'gen2node' in gen2node:
    >>>         gen2node = gen2node['gen2node']
    >>>     i = 1    
    >>>     for g, nodes in sorted(gen2node.items())[1:]:
    >>>         n = len(nodes)            
    >>>         if sorted(y[i:i+n]) != sorted(nodes):
    >>>             return False
    >>>         i+=n
    >>>     return True

    >>> x = timer(topological_sort, n = 10000)(dag, node)
    >>> y = timer(nx_topological_sort, n = 10000)(G, node)
    >>> assert compare_topological_sort_and_nx(x,y)

    Just to compare speeds, our topological sort is approx 10 times faster...

    >>> x = timer(topological_sort, n = 10000, time = True)(dag, node)
    >>> y = timer(nx_topological_sort, n = 10000, time = True)(G, node)
    >>> assert x * 10 < y   
    
    
    >>> DAGs['test'] = {}
    >>> dag = get_DAG('test')
    >>> for b in range(100):
    >>>     for a in range(b-3,b+3):
    >>>         add_edge('a%i'%a, 'b%i'%b, dag)
    >>> for c in range(100):
    >>>     for a in range(c-3,c+3):
    >>>         add_edge('a%i'%a, 'c%i'%c, dag)
    >>>     for b in range(c-3,c+3):
    >>>         add_edge('b%i'%b, 'c%i'%c, dag)
    >>> for d in range(100):
    >>>     for c in range(d-10,d+10):
    >>>         add_edge('c%i'%c, 'd%i'%d, dag)
    >>> for e in range(100):
    >>>     for c in range(e-10,e+10):
    >>>         add_edge('c%i'%c, 'e%i'%e, dag)
    >>>     for a in range(e-3,e+3):
    >>>         add_edge('a%i'%a, 'e%i'%e, dag)

    >>> import networkx as nx
    >>> G = nx.DiGraph()
    >>> for a in dag:
    >>>     for b in dag[a]:
    >>>         G.add_edge(a,b)

    >>> node = 'a50'

    >>> x = timer(topological_sort, n = 1000)(dag, node)
    >>> y = timer(nx_topological_sort, n = 1000)(G, node)
    >>> assert compare_topological_sort_and_nx(x,y)

    >>> x = timer(topological_sort, n = 1000, time = True)(dag, node)
    >>> y = timer(nx_topological_sort, n = 1000, time = True)(G, node)
    >>> assert x * 10 < y   
    """
    dag = get_DAG(dag)
    nodes = set(node) if isinstance(node, (list, set, dict_keys, dict_values)) else set([node])
    gen2node = {0 : nodes}
    node2gen = {node : 0 for node in nodes}
    gen = 0
    while len(gen2node[gen]):
        next_gen = gen + 1
        gen2node[next_gen] = set()
        for node in gen2node[gen].copy():
            if node in dag:
                children = dag[node]
                for child in children:                
                    if child in node2gen:
                        cgen = node2gen[child]
                        if cgen < next_gen:
                            gen2node[cgen].remove(child)
                            gen2node[next_gen].add(child)
                            node2gen[child] = next_gen
                    else:
                        gen2node[next_gen].add(child)
                        node2gen[child] = next_gen
        gen = gen + 1
    del gen2node[gen]
    return dict(gen2node = gen2node, node2gen = node2gen)
 

def descendants(dag, node, exc = None):
    """
    returns the descendants of the node (with node itself as first element) sorted topologically

    Parameters
    ----------
    dag : DAG
        A dag or a reference to a DAG
    node : a key
        A key to a node in the graph.
    exc: None or int or list of ints
        Most useful is to use exc = 0 to exclude the zeroth generation from future calculations

    Returns
    -------
    list
        list of keys for children of node in the graph.
    """
    gen = topological_sort(dag, node)['gen2node']
    if len(gen) == 0:
        return []
    else:
        excs = as_list(exc)
        if len(excs):
            nodes = [n for g, n in sorted(gen.items()) if g not in excs]
        else:
            nodes = gen.values()
        return sum(map(list,nodes), [])





def dag_inverse(dag = None):
    d = get_DAG(dag)
    g = {}
    for a in d:
        for b, value in d[a].items():
            if b not in g:
                g[b] = {}
            g[b][a] = value 
    get_GADs()[dag] = g
    return g


def add_edge(parent, child, dag = None, value = None):
    dag = get_DAG(dag)
    if parent not in dag:
        dag[parent] = {child : value}
    else:
        dag[parent].update({child : value})
    return dag

def del_edge(parent, child, dag = None):
    dag = get_DAG(dag)
    if parent in dag and child in dag[parent]:
        del dag[parent][child]
    return dag




from pyg_base._types import is_int, is_float, is_str, is_num, is_nan, is_ts, is_date
from pyg_base._as_list import as_list
from pyg_base._cache import cache
import datetime
import re
import pandas as pd
import numpy as np
from dateutil import parser
# from dateutil.relativedelta import relativedelta
from functools import reduce, partial
import dateutil as du
from dateutil import tz
from pytz import country_timezones
import pytz


TMIN = datetime.datetime(1900,1,1)
TMAX = datetime.datetime(2300,1,1)
microsecond = datetime.timedelta(microseconds = 1)
DAY = datetime.timedelta(days = 1)
iso = re.compile('^[0-9]{4}-[0-9]{2}-[0-9]{2}T')
ambiguity = re.compile('^[0-9]{1,2}[-/ .][0-9]{1,2}[-/ .][0-9]{4}')
futcodes = list('fghjkmnquvxz'.upper())
months = ['jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov', 'dec']
yyyymm = re.compile('^[0-9]{4}[-/ .][0-9]{1,2}$')
yyyymmm = re.compile('^[0-9]{4}[-/ .](%s)[a-z]*$'%('|'.join(months)), re.IGNORECASE)
wkdays = dict(mon = 0, tue = 1, wed = 2, thu = 3, fri = 4, sat = 5, sun = 6)
period = re.compile('^[-+]{0,1}[0-9]*[dbwmqyhnsDBWMQYHNS]$')


__all__ = ['dt','dt_bump', 'today', 'ymd', 'TMIN', 'TMAX', 'DAY', 'futcodes', 'dt2str', 'is_period', 'nth_weekday_of_month']

@cache
def tzones():
    x = {iso :country_timezones[iso] for iso in country_timezones}
    tzs = {iso: code[0] for iso, code in x.items() if len(code) == 1}
    for iso, code in x.items():
        tzs.update({c.split('/')[-1].replace('_', ' ').lower(): c for c in code})
    tzs  = {key : tz.gettz(value) for key, value in tzs.items()}
    t = datetime.datetime.now()
    tzs.update({v.tzname(t): v for v in tzs.values() if v is not None})
    tzs['WET'] = tzs['western european'] = tzs['MA'] # Malta
    tzs['EET'] = tzs['eastern european'] = tzs['GR']
    tzs['CET'] = tzs['cental european'] = tzs['AT'] # austria
    tzs['EST'] = tzs['eastern standard'] = tzs['new york']
    tzs['CST'] = tzs['central standard'] = tzs['chicago']
    tzs['greenwich mean time'] = tzs['GMT']
    return tzs

def as_tz(tzinfo):
    if is_str(tzinfo):
        zones = tzones()
        if tzinfo in zones:
            return zones[tzinfo]
        elif tzinfo.lower() in zones:
            return zones[tzinfo.lower()]
        elif tzinfo.upper() in zones:
            return zones[tzinfo.upper()]
        else:
            return pytz.timezone(tzinfo)
    else:
        return tzinfo

def tz_convert(t, tzinfo = None):
    """
    if tzinfo is None: drops all timezone data
    if tzinfo is not None: converts/localize the existing date to 
    

    Parameters
    ----------
    t : TYPE
        DESCRIPTION.
    tzinfo : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    if tzinfo is None:
        ## we remove tzinfo
        if isinstance(t, (pd.DataFrame, pd.Series)):
            if t.index.tz is None:
                return t
            else:
                res = t.copy()
                res.index = [date.replace(tzinfo = None) for date in res.index]
                return res
        elif is_date(t):
            return t if t.tzinfo is None else t.replace(tzinfo = None)
        elif isinstance(t, list):
            return type(t)([tz_convert(i, tzinfo) for i in t])
        elif isinstance(t, dict):
            return type(t)({key : tz_convert(value, tzinfo) for key, value in t.items()})
        else:
            return dt(t, tzinfo = tzinfo)
    if isinstance(t, list):
        return type(t)([tz_convert(i, tzinfo) for i in t])
    elif isinstance(t, dict):
        return type(t)({key : tz_convert(value, tzinfo) for key, value in t.items()})
    if isinstance(t, (pd.DataFrame, pd.Series)):
        res = t.copy()
        if t.index.tz is None:
            res.index = res.index.tz_localize(as_tz(tzinfo))
        else:
            res.index = res.index.tz_convert(as_tz(tzinfo))
        return res
    elif is_date(t):
        return t.astimezone(as_tz(tzinfo))
    else:
        return dt(t, tzinfo)

def today(date = None):
    now = date or datetime.datetime.now()
    return datetime.datetime(now.year, now.month, now.day)


def month(m):
    """
    converts m into a valid month
    
    :Examples:
    ---------------
    >>> assert month(3) == 3
    >>> assert month('h') == 3
    >>> assert month('march') == 3

    :Parameters:
    ----------------
    m : str/int
        DESCRIPTION.    

    Raises
    ------
    ValueError
        DESCRIPTION.

    :Returns:
    -------
    int
        month in integer 1...12

    """
    if is_int(m):
        return m
    elif is_float(m) and int(m) == m:
        return int(m)
    elif is_str(m):
        if len(m) == 1 :
            return futcodes.index(m.upper()) + 1
        else:
            return months.index(m[:3].lower()) + 1
    else:
        raise ValueError('do not understand month %s' % m)

def ym(y,m):
    """
    converts a y,m into actual year-month
    
    :Example:
    --------------
    >>> assert ym(2000, -1) == (1999,11)
    >>> assert ym(2000, 0) == (1999,12)
    >>> assert ym(2000, 1) == (2000,1)
    >>> assert ym(2000, 'h') == (2000,3)
    >>> assert ym(2000, 'March') == (2000,3)
    ...
    >>> assert ym(2000, 12) == (2000,12)
    >>> assert ym(2000, 13) == (2001,1)
    
    """
    y = int(y) if is_float(y) and int(y) == y else y
    m = month(m)
    y += (m-1) // 12
    m = 1 + ((m-1) % 12)
    return (y, m)

def _ymd(y,m,d):
    """
    returns a date from year, month, day.
    
    WARNING: Handles date arithmetic not quite how you expect it

    :Parameters:
    ----------------
    y : int
        year.
    m : int
        month offsets from 0.
    d : int
        day offsets from 0 

    :Returns:
    -------
    date
        calculates dates.

    """
    if d>1500 and d<3000 and y>0 and y<32 and m>0 and m<13:
        y,d = d,y
    y,m = ym(y,m)
    d = int(d) if is_float(d) and int(d) == d else d
    return datetime.datetime(y,m,1) + (d-1) * DAY

def num2dt(n):
    i = int(n); f = datetime.timedelta(n - i)
    if i<=1500:
        return today() + i * DAY + f
    elif i<=3000:
        return datetime.datetime(i, 1, 1) + f
    elif i < 300000:
        return datetime.datetime.fromordinal(i + 693594) + f
    elif i<1095000:
        return datetime.datetime.fromordinal(i) + f
    elif i>10000101 and i<30001231:
        y = i // 10000
        m = (i % 10000) // 100
        d = i % 100
        return _ymd(y,m,d) + f
    else:
        return datetime.datetime.utcfromtimestamp(n)

# def int2dt(i):
#     if i<=1500:
#         return today() + i * day
#     elif i<=3000:
#         return datetime.datetime(i, 1, 1)
#     elif i < 300000:
#         return datetime.datetime.fromordinal(i + 693594)
#     elif i<1095000:
#         return datetime.datetime.fromordinal(i)
#     elif i>10000101 and i<30001231:
#         y = i // 10000
#         m = (i % 10000) // 100
#         d = i % 100
#         return ymd(y,m,d)
#     else:
#         return datetime.datetime.utcfromtimestamp(i)
    
def np2dt(t):
    """
    >>> d = datetime.datetime(2000,1,1,20,30,40,55)
    >>> t = np.datetime64(d)    
    >>> assert dt(t) == d
    >>> t = np.datetime64(d).astype('datetime64[D]')
    >>> assert np2dt(t) == dt(d)
    >>> t = np.datetime64(d).astype('datetime64[m]')
    >>> assert dt(t) == datetime.datetime(2000,1,1,20,30)
    >>> t = np.datetime64(d).astype('datetime64[ms]')
    >>> assert dt(t) == datetime.datetime(2000,1,1,20,30,40)
    >>> t = np.datetime64(d).astype('datetime64[ns]')
    >>> assert dt(t) == d
    
    Parameters
    ----------
    t : numpy.datetime64 format
        time.

    Returns
    -------
    datetime
        datetime object.
    """
    
    res = t.astype(datetime.datetime)
    if isinstance(res, datetime.datetime):
        return res
    elif isinstance(res, datetime.date): # [D] format
        return datetime.datetime(res.year, res.month, res.day)
    elif is_int(res): ## [ns] format
        return pd.Timestamp(t)
    return res

        
def uk2dt(t, tzinfo = None):
    if t in ('', 'null'):
        return None
    res = parser.parse(t)
    if ambiguity.search(t) is not None:
        if res.day<13:
            res = dt(res.year, res.day, res.month, res.hour, res.minute, res.second, res.microsecond)
        elif int(t[:2].replace('-','').replace('/',''))!=res.day:
            raise ValueError('date %s is not in UK date format'%t)
    elif yyyymm.search(t) is not None or yyyymmm.search(t) is not None:
        res = datetime.datetime(res.year, res.month, 1)
    return tz_convert(res, tzinfo)


def us2dt(t, tzinfo = None):
    if t in ('', 'null'):
        return None
    res = parser.parse(t)
    if ambiguity.search(t) is not None and res.month != int(t[:2].replace('-','').replace('/','')):
        raise ValueError('the date is not in US format')
    if yyyymm.search(t) is not None or yyyymmm.search(t) is not None:
        res = datetime.datetime(res.year, res.month, 1)
    return tz_convert(res, tzinfo)

def none2dt(none = datetime.datetime.now):
    if callable(none):
        return none()
    else:
        return none

def is_period(bump):
    return is_str(bump) and period.search(bump) is not None    

def is_bump(bump):
    return is_period(bump) or (is_int(bump) and bump<1500) or isinstance(bump, (datetime.timedelta, du.relativedelta.relativedelta))

def dt_bump(t, *bumps):
    """
    :Example:
    ---------
    >>> from pyg import *
    >>> t  = pd.Series([1,2,3], drange(dt(2000,1,1),2))
    >>> assert eq(dt_bump(t, 1), pd.Series([1,2,3], drange(dt(2000,1,2),2)))

        
    """
    bumps = as_list(bumps)
    if is_ts(t):
        res = t.copy()
        res.index = [dt_bump(i, *bumps) for i in res.index]
        return res
    t = t if isinstance(t, datetime.datetime) else dt(t)
    for bump in bumps:
        if is_int(bump):
            t = t + DAY * bump
        elif isinstance(bump, (datetime.timedelta, du.relativedelta.relativedelta)):
            t = t + bump
        elif is_str(bump):
            bump = bump.lower()
            if not is_period(bump):
                try:
                    t = tz_convert(t, bump)
                except Exception:
                    raise ValueError('%s is not a period I know...'%bump)
            if bump.endswith('d'):
                t = t + DAY * int(bump[:-1])
            elif bump.endswith('w'):
                t  = t + DAY * (7 * int(bump[:-1]))
            elif bump.endswith('m'):
                t = _ymd(t.year, t.month + int(bump[:-1]), t.day)
            elif bump.endswith('q'):
                t = _ymd(t.year, t.month + 3  * int(bump[:-1]), t.day)
            elif bump.endswith('y'):
                t = _ymd(t.year+int(bump[:-1]), t.month, t.day)
            elif bump.endswith('h'):
                return t + datetime.timedelta(hours = int(bump[:-1]))
            elif bump.endswith('n'):
                return t + datetime.timedelta(minutes = int(bump[:-1]))
            elif bump.endswith('s'):
                return t + datetime.timedelta(seconds = int(bump[:-1]))
            elif bump.endswith('b'):
                bdays = int(bump[:-1])
                wday = t.weekday()
                if wday>4:
                    t = t + (7-wday) * DAY
                    wday = 0
                w = bdays // 5
                d = bdays - w * 5
                t = t + DAY * (7*w)
                if wday + d > 4:
                    d+=2
                t += DAY * d
        else:
            t = t + bump
    return t

def dt(*args, dialect = 'uk', none = datetime.datetime.now, tzinfo = None):
    """
    A more generic constructor for datetime.datetime. 
    
    :Example: Simple construction
    --------------
    >>> assert dt(2000,1 ,1) == datetime.datetime(2000, 1, 1, 0, 0) # name of month
    >>> assert dt(2000,'jan',1) == datetime.datetime(2000, 1, 1, 0, 0) # name of month
    >>> assert dt(2000,'f',1) == datetime.datetime(2000, 1, 1, 0, 0) # future month code
    >>> assert dt('01-02-2002') == datetime.datetime(2002, 2, 1)
    >>> assert dt('01-02-2002', dialect = 'US') == datetime.datetime(2002, 1, 2)
    >>> assert dt('01 March 2002') == datetime.datetime(2002, 3, 1)
    >>> assert dt('01 March 2002', dialect = 'US') == datetime.datetime(2002, 3, 1)
    >>> assert dt('01 March 2002 10:20:30') == datetime.datetime(2002, 3, 1, 10, 20, 30)

    >>> assert dt(20020301) == datetime.datetime(2002, 3, 1)
    >>> assert dt(37316) == datetime.datetime(2002, 3, 1) # excel date
    >>> assert dt(730180) == datetime.datetime(2000,3,1) # ordinal for 1/3/2000
    >>> assert dt(2000,3,1).timestamp() == 951868800.0
    >>> assert dt(951868800.0) == datetime.datetime(2000,3,1) # utc timestamp
    >>> assert dt(np.datetime64(dt(2000,3,1))) == dt(2000,3,1) ## numpy.datetime64 object

    >>> assert dt(2000) == datetime.datetime(2000,1,1)
    >>> assert dt(2000,3) == datetime.datetime(2000,3,1)
    >>> assert dt(2000,3, 1) == datetime.datetime(2000,3,1)
    >>> assert dt(2000,3, 1, 10,20,30) == datetime.datetime(2000,3,1,10,20,30)
    >>> assert dt(2000,'march', 1) == datetime.datetime(2000,3,1)
    >>> assert dt(2000,'h', 1) == datetime.datetime(2000,3,1) # future codes


    :Example: date as offset from today
    -----------------------------------
    >>> today = dt(0); 
    >>> import datetime
    >>> day = datetime.timedelta(1)
    >>> assert dt(-3) == today - 3 * day
    >>> assert dt('-10b') == today - 14 * day
    
    :Example: datetime arithmetic:
    -----------------------------------------------
    dt has an interesting logic in implementing datetime arithmentic: 
        
        - day and month parameters can be negative or bigger than the days of month
        - dt() will roll back/forward from the date which is valid
    
    >>> assert dt(2000,4,1) == datetime.datetime(2000, 4, 1, 0, 0)
    >>> assert dt(2000,4,0) == datetime.datetime(2000, 3, 31, 0, 0) # a day before dt(2000,4,1)

    and rolling back months:
        
    >>> assert dt(2000,0,1) == datetime.datetime(1999, 12, 1, 0, 0) # a month before dt(2000,1,1)
    >>> assert dt(2000,13,1) == datetime.datetime(2001, 1, 1, 0, 0) # a month after dt(2000,12,1)
    
    This may feel unnatural at first, but does allow for much nicer code, e.g.:
    [dt(2000,i,1) for i in range(-10,10)]

    :Parameters:
    ----------------
    *args : str, int or dates
        argument to be converted into dates
    dialect : str, optional
        parsing of 01/02/2020 is it 1st Feb or 2nd Jan? The default is 'uk', i.e. dd/mm/yyyy
    none : callable, optional
        What is dt()? The default is datetime.datetime.now()
    
    """
    args = as_list(args)
    if len(args) == 0:
        res = none() if callable(none) else none
        return tz_convert(res, tzinfo)
    t = args[0]
    if isinstance(t, np.datetime64):
        t = tz_convert(np2dt(t), tzinfo)
    elif isinstance(t, datetime.date) and not isinstance(t, datetime.datetime):
        t = datetime.datetime(t.year, t.month, t.day, tzinfo = as_tz(tzinfo))
    if isinstance(t, datetime.datetime):
        return reduce(dt_bump, args[1:], tz_convert(t, tzinfo))
    if len(args) == 1:
        if t is None:
            return tz_convert(none(), tzinfo) if callable(none) else none
        elif is_num(t):
            if is_nan(t):
                return tz_convert(none(), tzinfo) if callable(none) else none
            else:                
                return tz_convert(num2dt(t), tzinfo)
        elif is_bump(t):
            return dt_bump(dt(0, tzinfo = tzinfo), t)
        elif is_str(t):
            return uk2dt(t, tzinfo = tzinfo) if dialect == 'uk' else us2dt(t, tzinfo = tzinfo)
                # return int2dt(int(t)) + datetime.timedelta(float(t) % 1)
        else:
            raise ValueError('date format unrecognised %s'%t)
    elif len(args) == 2:
        y,m = ym(*args)
        return datetime.datetime(y,m,1, tzinfo = as_tz(tzinfo))
    y,m,d = args[:3]
    t = _ymd(y,m,d)
    if len(args) == 4 and is_str(args[3]):
        return tz_convert(nth_weekday_of_month(*args), tzinfo)
    if len(args) > 3:
        args = [int(a) for a in args[3:]] + [0,0,0]
        return tz_convert(t + datetime.timedelta(hours = args[0], minutes = args[1], seconds = args[2]), tzinfo)
    else:
        return tz_convert(t, tzinfo)


def nth_weekday_of_month(y, m, n, w):
    """
    y = 2020; m = 2
    assert nth_weekday_of_month(y, m, -1, 'sat') == dt(2020, 2, 29)
    assert nth_weekday_of_month(y, m, -2, 'sat') == dt(2020, 2, 22)
    assert nth_weekday_of_month(y, m, 1, 'sat') == dt(2020, 2, 1)
    assert nth_weekday_of_month(y, m, 1, 'sun') == dt(2020, 2, 2)
    assert nth_weekday_of_month(y, m, 1, 'monday') == dt(2020, 2, 3)
    assert nth_weekday_of_month(y, 'G', 3, 'sat') == dt(2020, 2, 15)
    assert nth_weekday_of_month(y, 'G', 3, 'sun') == dt(2020, 2, 16)
    assert nth_weekday_of_month(y, 'G', 3, 'monday') == dt(2020, 2, 17)
    """
    if n < 0 :
        return nth_weekday_of_month(y, m+1, 1, w) + datetime.timedelta(7 * n)
    t = dt(y, m , 1)
    bump = wkdays[w[:3].lower()] - t.weekday()
    if bump < 0:
        bump = bump + 7
    bump = bump + (n-1) * 7 
    res = t + datetime.timedelta(bump)
    return res

def ymd(*args, dialect = 'uk', none = datetime.datetime.now, tzinfo = None):
    """
    just like dt() but always returns date only (year/month/date) without fractions.
    see dt() for full documentation
    
    Returns
    -------
    datetime.datetime

    """
    t = dt(*args, dialect = dialect , none = none, tzinfo = tzinfo)
    return datetime.datetime(t.year, t.month, t.day, tzinfo = t.tzinfo)


ndt = partial(dt, none = None)
ndt.now = datetime.datetime.now
ndt.today = today
ndt.timedelta = datetime.timedelta

dt.now = datetime.datetime.now
dt.today = today
dt.timedelta = datetime.timedelta


def _date_format(fmt = None):
    """
    convenience function to allow quick format string

    :Parameters:
    ----------------
    fmt : str
        DESCRIPTION. The default is None.

    :Returns:
    -------
    fmt : str
        A valid format string for dates.
        
    :Examples:
    ---------------
    _date_format('Y/M')

    """
    if fmt is None: 
        fmt = ''
    if is_str(fmt) and len(fmt)<=1:
        fmt = fmt.join(['%Y', '%m', '%d'])
    if '%' not in fmt:
        fmt = ''.join(['%'+x if x in 'aAwdbBmyYHIpMSfzZjUWcXx' else x for x in fmt])
    return fmt

def dt2str(t, fmt = None):
    """
    converts a date into a string format. fmt supports the formats as specified by datetime().strftime() but also simplifies it
    
    :Parameters:
    -----------------
    t : date
    fmt : a format string
    
        if fmt is None:
            If the date has no intra-day component, by defauult will go to yyyymmdd format.
            If the date has intra-day component, will go to iso-format.
    
    :Examples:
    ---------------
    >>> assert dt2str(2000)  == '20000101'
    >>> assert dt2str(2000, 'Ymd')  == '20000101'
    >>> assert dt2str(2000, '-') == '2000-01-01'
    >>> assert dt2str(2000, '%B %Y') == 'January 2000'
    >>> assert dt2str(2000, 'B Y') == 'January 2000'
    >>> assert dt2str(2000, 'iso') == '2000-01-01T00:00:00'
    >>> t = datetime.datetime(2000,1,10,20,30,40,50)
    >>> assert dt2str(t) == '2000-01-10T20:30:40.000050'
    """
    t = t if isinstance(t, datetime.datetime) else dt(t)
    if fmt is None:
        if t == today(t):
            return t.strftime('%Y%m%d')
        else:
            return t.isoformat()
    fmt = _date_format(fmt)
    if fmt.lower() == 'iso':
        return t.isoformat()
    else:
        return t.strftime(fmt)

    



import pandas as pd
import numpy as np
from _collections_abc import Iterable, dict_keys, dict_values
import datetime
from functools import partial

from pyg_base._logger import logger

__all__ = ['is_series', 'is_df', 'is_pd', 'is_arr', 'is_list', 'is_len', 'is_tuple', 'is_int', 'is_float', 'is_num', 'is_bool', 'is_str', 'is_nan', 'is_none', 'is_dict', 'is_iterable', 'is_date', 'is_ts']
__all__ += ['is_pds', 'is_arrs', 'is_ints', 'is_floats', 'is_nums', 'is_bools', 'is_strs', 'is_nans', 'is_nones', 'is_dicts', 'is_iterables', 'is_dates', 'is_tss', 'is_lists', 'is_tuples']
__all__ += ['nan2none', 'NoneType']
__all__ = sorted(__all__)
NoneType = type(None)

def is_series(value):
    """ is value a pd.Series"""
    return isinstance(value, pd.Series)

def is_df(value):
    """ is value a pd.DataFrame"""
    return isinstance(value, pd.DataFrame)

    
def is_pd(value):
    """ is value a pd.DataFrame/pd.Series"""
    return isinstance(value, (pd.Series, pd.DataFrame))

def is_arr(value):
    """ is value a numpy array of non-zero-size"""
    return isinstance(value, np.ndarray) and len(value.shape) > 0

is_array = is_arr

def is_int(value):
    """ is value an int, or any variant of np.intN type"""
    return isinstance(value, (int, np.int64, np.int32, np.int16, np.int8))

def is_float(value):
    """ is value an float, or any variant of np.float """
    return isinstance(value, (float, np.float16, np.float32, np.float64))

def is_num(value):
    """ is _int(value) or is_float(value)"""
    return is_int(value) or is_float(value)

def is_bool(value):
    """ is value a Bool, or a np.bool_ type"""
    return isinstance(value, (bool, np.bool_))

def is_str(value):
    """ is value a str, or a np.str_ type"""
    return isinstance(value, (str, np.str_))

def is_nan(value):
    """ is value a nan or an inf. Unlike np.isnan, works for non numeric"""
    return is_float(value) and (np.isnan(value) or np.isinf(value))

def is_none(value):
    """ is value None"""
    return value is None

def is_dict(value):
    """ is value a dict"""
    return isinstance(value, dict)

def is_zero_len(value):
    """ is value of zero length (or has no len at all)"""
    return getattr(value ,'__len__', lambda : 0)() == 0

def is_len(value):
    """ is value of zero length (or has no len at all)"""
    return getattr(value ,'__len__', lambda : 0)() > 0

def is_iterable(value):
    """is value Iterable excluding a string"""
    return isinstance(value, Iterable) and not is_str(value)


def nan2none(value):
    """convert np.nan/np.inf to None"""
    return None if is_nan(value) else value

def is_list(value):
    """is value a list"""
    return isinstance(value, list)

def is_tuple(value):
    """is value a tuple"""
    return isinstance(value, tuple)

def is_listable(value):
    """is value a tuple, list or np.ndarray"""
    return is_iterable(value) and not isinstance(value, (dict, pd.Series, pd.DataFrame))

def is_date(value):
    """is value a date type: either datetime.date, datetime.datetime or np.datetime64"""
    return isinstance(value, (datetime.date, datetime.datetime, np.datetime64))


def is_primitive(value):
    return value is None or is_bool(value) or is_num(value) or is_date(value) or is_str(value)


def is_ts(value):
    """is value a pandas datafrome whch is indexed by datetimes"""
    if is_pd(value):
        if len(value):
            t0, t1 = value.index[0], value.index[-1]
            if is_date(t0) and is_date(t1):
                if t0>t1:
                    logger.warning('WARN: unsorted timeseries %s...%s'%(t0, t1))
                return True
            else:
                return False
        else:
            return True
    else:
        return False
                    
def is_one_or_many(value, check):
    return check(value) or (is_iterable(value) and is_len(value) and check(value[0]) and min([check(v) for v in value]))


is_pds = partial(is_one_or_many, check = is_pd) 

is_arrs = partial(is_one_or_many, check = is_arr) 

is_ints = partial(is_one_or_many, check = is_int) 

is_floats = partial(is_one_or_many, check = is_float) 

is_nums = partial(is_one_or_many, check = is_num) 

def is_strs(value):
    return is_str(value) or (isinstance(value, (list, tuple, dict_keys, dict_values)) and is_len(value) and min([is_str(v) for v in value]))

is_nans = partial(is_one_or_many, check = is_nan) 

is_dicts = partial(is_one_or_many, check = is_dict) 

def is_lists(value):
    check = is_list
    return is_iterable(value) and is_len(value) and check(value[0]) and min([check(v) for v in value])

is_bools = partial(is_one_or_many, check = is_bool) 

is_nones = partial(is_one_or_many, check = is_none) 

is_dates = partial(is_one_or_many, check = is_date)

is_iterables = partial(is_one_or_many, check = is_iterable) 

is_tss = partial(is_one_or_many, check = is_ts) 

is_tuples = partial(is_one_or_many, check = is_tuple) 



from pyg_base._dictattr import dictattr
from pyg_base._decorators import kwargs_support, try_none
from pyg_base._as_list import as_list, as_tuple
from pyg_base._types import is_str
from pyg_base._inspect import getargs
from pyg_base._eq import in_
from copy import copy



__all__ = ['Dict', 'dict_invert', 'items_to_tree', 'tree_items', 'tree_keys', 'tree_values', 'tree_update', 'tree_setitem', 'tree_getitem', 'tree_get']

    
class Dict(dictattr):
    """
    Dict extends dictattr to allow access to *functions* of members
    
    :Example:
    --------------
    >>> from pyg import *
    >>> d = Dict(a = 1, b=2)
    >>> assert d[lambda a, b: a+b] == 3
    >>> assert d['a','b', lambda a,b: a+b] == [1,2,3]
    
    Dict is also callable where the key-value is used to add/update current members        
    
    :Example:
    -------
    >>> from pyg import *
    >>> d = Dict(a = 1, b=2)
    >>> assert d(c = 3) == Dict(a = 1, b = 2, c = 3)
    >>> assert d(c = lambda a,b: a+b) == Dict(a = 1, b = 2, c = 3)

    
    >>> assert d(c = 3) == Dict(a = 1, b = 2) + Dict(c = 3)
    >>> assert Dict(a = 1)(b = lambda a: a+1)(c = lambda a,b: a+b) == Dict(a = 1,b = 2,c = 3)
    """
    def __getitem__(self, value):
        if callable(value):
            return self.apply(value)
        else:
            return super(Dict, self).__getitem__(value)
    
    def apply(self, function):
        return kwargs_support(function)(**self)
    
    def __call__(self, **kwargs):
        res = self.copy()
        res.update({key : value for key, value in kwargs.items() if not callable(value)})
        callables = {key: value for key, value in kwargs.items() if callable(value)}
        while len(callables) > 1:        
            keys = set(callables.keys())        
            independent = {key: value for key, value in callables.items() if len(keys & set(getargs(value))) == 0}
            if len(independent) == 0:
                raise ValueError('circular function calling')
            else:
                for key, value in independent.items():
                    res[key] = res[value]
                callables = {key: value for key, value in callables.items() if not key in independent}
        for key, value in callables.items():
            res[key] = res[value]
        return res
    
    def __add__(self, other):
        return tree_update(self, other)
    
    def do(self, function, *keys):
        """
        applies a function(s) on multiple keys at the same time

        :Parameters:
        ----------------
        function : callable or list of callables
            function to be applied per column
        *keys : string/list of strings
            list of columns to be applied. If missing, applied to all columns

        :Returns:
        -------
        res : Dict

        :Example:
        --------------
        >>> from pyg import *
        >>> d = Dict(name = 'adam', surname = 'atkins')
        >>> assert d.do(proper) == Dict(name = 'Adam', surname = 'Atkins')

        :Example: using another key in the calculation
        -------
        >>> from pyg import *
        >>> d = Dict(a = 1, b = 5, denominator = 10)
        >>> d = d.do(lambda value, denominator: value/denominator, 'a', 'b')
        >>> assert d == Dict(a = 0.1, b = 0.5, denominator = 10)
        
        """
        res = self.copy()
        keys = as_list(keys)
        if len(keys)  == 0:
            keys = self.keys()
        for key in keys:
            for f in as_list(function):
                args = as_list(try_none(getargs)(f))
                res[key] = f(res[key], **{k : v for k, v in res.items() if k in args[1:]})
        return res


def dict_invert(d):
    """
    inverts a dict to create value -> keys

    :Example:
    --------------
    >>> d = dict(a = 1, b = 2, c = 1, d = 2, e = 4)
    >>> assert dict_invert(d) == {1: ['a', 'c'], 2: ['b', 'd'], 4: ['e']}
    
    
    If you think your data is unique:
    >>> d = dict(a = 1, b = 2, c = 3)
    >>> assert dict_invert(d).do(first) ==  {1: 'a', 2: 'b', 3: 'c'}

    :Parameters:
    ----------------
    d : dict
        a dict whose key-to-value.

    :Returns:
    -------
    dict of values to [lists]

    """
    res = Dict()
    for key, value in d.items():
        res[value] = res.get(value, []) + [key]
    return res


def _tree_types(types):
    return (dict, Dict, dictattr) if types is None else as_tuple(types)


def _tree_setitem(tree, item, base, ignore, types):
    if len(item)<2:
        raise ValueError('node item too short %s'%item)
    res = tree
    for key in item[:-2]:
        if key not in res or not isinstance(res[key], types):
            res[key] = base()
        res = res[key]
    if item[-2] in res and in_(item[-1], ignore):
        return
    else:
        res[item[-2]] = item[-1]

def tree_setitem(tree, key, value, ignore = None, types = None):
    """
    sets an item of a tree
    
    :Parameters:
    ----------
    tree : tree (dicts of dict)
    key : a dot-separated string or a tuple of values
        the branch to hang value on
    value : object
        the leaf at the end of the branch
    ignore : None or list, optional
        what values of leaf will be ignored and not overwrite existing data. The default is None.
    types : types, optional
        As we go down the tree, when do we stop and say: what we have is a leaf already?

    :Example:
    ---------
    >>> tree = dict()
    >>> tree_setitem(tree, 'a', 1)
    >>> assert tree == dict(a = 1)
    >>> tree_setitem(tree, 'b.c', 2)
    >>> assert tree == {'a': 1, 'b': {'c': 2}}
    >>> tree_setitem(tree, ('b','c','d'), 2)
    >>> tree_setitem(tree, ('b','c','e'), 3)
    >>> assert tree == {'a': 1, 'b': {'c': {'d': 2, 'e': 3}}}

    :Example: types
    ---------------
    >>> from pyg import *
    >>> tree = dict(mycell = cell(lambda a, b: a * b, b = 2, a = cell(lambda x: x**2, x = cell(lambda y: y*3))))
    >>> # We are missing y....
    >>> tree_setitem(tree, 'mycell.a.x.y', 3, types = (dict,cell)) ## drill into cell
    >>> assert tree['mycell'].a.x.y == 3
    >>> tree_setitem(tree, 'mycell.a.x.y', 1) ## stop when you hit cell
    >>> assert tree['mycell'].a.x == dict(y = 1)

    Returns
    -------
    None.

    """
    
    types = _tree_types(types)
    ignore = as_list(ignore)
    base = type(tree)
    if is_str(key):
        key = key.split('.')
    elif isinstance(key, tuple):
        key = list(key)
    _tree_setitem(tree, key + [value], base = base, ignore = ignore, types = types)

def tree_getitem(tree, item):
    """
    :Example:
    ---------
    >>> tree = dict(person = dict(name = 'james', surname = 'blackburn', age = 30))
    >>> assert tree_getitem(tree, 'person.name') == 'james'
    >>> assert tree_getitem(tree, 'person.age') == 30
    >>> assert tree_getitem(tree, ['person','surname']) == 'blackburn'
    """
    items = item.split('.') if isinstance(item, str) else as_list(item)
    res = tree
    for i in items:
        res = res[i]
    return res
        
def tree_get(tree, item, default = None):
    """
    :Example:
    ---------
    >>> tree = dict(person = dict(name = 'james', surname = 'blackburn', age = 30))
    >>> assert tree_get(tree, 'person.salary') is None
    >>> assert tree_get(tree, 'person.age') == 30
    >>> assert tree_get(tree, ['person','surname', 'other'], 999) == 999
    """
    items = item.split('.') if isinstance(item, str) else as_list(item)
    res = tree
    for i in items:
        if isinstance(res, dict) and i in res:
            res = res[i]
        else:
            return default
    return res
    


def tree_items(tree, types = None):
    """
    An extension of dict.items(), returning a list of tuples but of varying length, each a branch of a tree

    :Parameters:
    ----------------
    tree : dict of dicts
        a tree of data.
    types : dict or a list of dict-types, optional
        The types that we consider as 'branches' of the tree. Default is (dict, Dict, dictattr).

    :Returns:
    -------
    a list of tuples
        these are an extension of dict.items() and are of varying length
        
    :Example:
    --------------
    >>> school = dict(pupils = dict(id1 = dict(name = 'james', surname = 'maxwell', gender = 'm'),
                          id2 = dict(name = 'adam', surname = 'smith', gender = 'm'),
                          id3 = dict(name = 'michell', surname = 'obama', gender = 'f'),
                          ),
            teachers = dict(math = dict(name = 'albert', surname = 'einstein', grade = 3),
                            english = dict(name = 'william', surname = 'shakespeare', grade = 3),
                            physics = dict(name = 'richard', surname = 'feyman', grade = 4)
                            ))

    >>> items = tree_items(school)
    >>> items 
    
    >>> [('pupils', 'id1', 'name', 'james'),
    >>>  ('pupils', 'id1', 'surname', 'maxwell'),
    >>>  ('pupils', 'id1', 'gender', 'm'),
    >>>  ('pupils', 'id2', 'name', 'adam'),
    >>>  ('pupils', 'id2', 'surname', 'smith'),
    >>>  ('pupils', 'id2', 'gender', 'm'),
    >>>  ('pupils', 'id3', 'name', 'michell'),
    >>>  ('pupils', 'id3', 'surname', 'obama'),
    >>>  ('pupils', 'id3', 'gender', 'f'),
    >>>  ('teachers', 'math', 'name', 'albert'),
    >>>  ('teachers', 'math', 'surname', 'einstein'),
    >>>  ('teachers', 'math', 'grade', 3),
    >>>  ('teachers', 'english', 'name', 'william'),
    >>>  ('teachers', 'english', 'surname', 'shakespeare'),
    >>>  ('teachers', 'english', 'grade', 3),
    >>>  ('teachers', 'physics', 'name', 'richard'),
    >>>  ('teachers', 'physics', 'surname', 'feyman'),
    >>>  ('teachers', 'physics', 'grade', 4)]
    
    #To reverse this, we call:
        
    >>> assert items_to_tree(items) == school
    
    """
    types = (dict, Dict, dictattr) if types is None else as_list(types)
    if type(tree) in types:
        return sum([[(key,) + item for item in tree_items(tree[key], types)] for key in tree], [])
    else: # this is a leaf
        return [(tree,)]
    
def tree_keys(tree, types = None):
    """
    returns the keys (branches) of a tree as a list of of tuples

    :Example:
    ---------
    >>> tree = dict(a = 1, b = dict(c = 2, d = 3, e = dict(f = 4)))
    >>> assert tree_keys(tree) == [('a',), ('b', 'c'), ('b', 'd'), ('b', 'e', 'f')]

    :Parameters:
    ----------
    tree : tree (dict of dicts)
    types : types of dicts, optional

    """
    types = _tree_types(types)
    if type(tree) in types:
        return sum([[(key,) + item for item in tree_keys(tree[key], types)] for key in tree], [])
    else: # this is a leaf
        return [()]
    
def tree_values(tree, types = None):
    """
    returns the values (leaf) of a tree (a collection of tuples)

    :Example:
    ---------
    >>> tree = dict(a = 1, b = dict(c = 2, d = 3, e = dict(f = 4)))
    >>> assert tree_values(tree) == [1,2,3,4]

    :Parameters:
    ----------
    tree : tree (dict of dicts)
    types : types of dicts, optional

    """
    types = _tree_types(types)
    if type(tree) in types:
        return sum([[item for item in tree_values(tree[key], types)] for key in tree], [])
    else: # this is a leaf
        return [tree]


    
def items_to_tree(items, tree = None, raise_if_duplicate = True, ignore = None, types = None):
    """
    converts **items** to branches of a tree. 
    If an original **tree** is provided, hang the additional branches on the existing tree
    If **ignore** is provided as a list of values, will not overwrite branches with last value (the leaf) in these values
    
    :Example:
    --------------
    >>> items = [('cambridge', 'smith', 'economics',),
             ('cambridge', 'keynes', 'economics'), 
             ('cambridge', 'lyons',  'maths'),
             ('cambridge', 'maxwell', 'maths'),
             ('oxford', 'penrose', 'maths'),
             ]
    
    >>> tree = items_to_tree(items)
    >>> print(tree_repr(tree))
    
    >>> cambridge:
    >>>     smith:
    >>>         economics
    >>>     keynes:
    >>>         economics
    >>>     lyons:
    >>>         maths
    >>>     maxwell:
    >>>         maths
    >>> oxford:
    >>>     {'penrose': 'maths'}
    
    We can add to tree:
    
    
    :Parameters:
    ----------------
    items : list of tuples, 
        items are just like dict items, only longer, 
    tree : tree, optional
        a pre-existing tree of trees. The default is None.
    raise_if_duplicate : TYPE, optional
        DESCRIPTION. The default is True.
    ignore : list, optional
        list of values that when over-writing an existing tree, should ignore. The default is None.


    :Example: using ignore
    ----------------------
    >>> tree = dict(a = 1, b = 'keep_old_value')
    >>> update = dict(a = 'valid_new_value', b = None, c = None)
    >>> tree_update(tree, update, ignore = [None])
    >>> {'a': valid_new_value, 'b': 'keep_old_value', 'c': None}

    * a is over-ridden as the new value is valid
    * b is not over-ridden since the update b = None is considereed invalid
    * c is added as it did not exist before, even though c = None is invalid value

    :Returns:
    -------
    tree : dict of dicts

    """
    types = _tree_types(types)
    if raise_if_duplicate and len(set([tuple(node[:-1]) for node in items])) < len(items):
        raise ValueError('items are not unique and will overwrite each other')
    if tree is None:
        tree = dictattr()
    else:
        tree = copy(tree)
    base = type(tree)
    ignore = as_list(ignore)
    for item in items:
        _tree_setitem(tree, item, base, ignore, types)
    return tree

def tree_update(tree, update, types = (dict, Dict, dictattr), ignore = None):
    """
    equivalent to dict.update() except: not in-place and also updates further down the tree
    
    :Example:
    --------------
    >>> ranking = dict(cambridge = dict(trinity = 1, stjohns = 2, christ = 3), 
                oxford = dict(trinity = 1, jesus = 2, magdalene = 3))
    >>> new_ranking = dict(oxford = dict(wolfson = 3, magdalene = 4))

    >>> print(tree_repr(tree_update(ranking, new_ranking)))

    >>> cambridge:
    >>>     {'trinity': 1, 'stjohns': 2, 'christ': 3}
    >>> oxford:
    >>>     {'trinity': 1, 'jesus': 2, 'magdalene': 4, 'wolfson': 3}

    Note how values for magdalene in Oxford were overwritten even though they are further down the tree

    :Example: using ignore
    -----------------------
    >>> update = dict(a = None, b = np.nan, c = 0)
    >>> tree = dict(a = 1, b = 2, c = 3)
    >>> assert tree_update(tree, update) == update
    >>> assert tree_update(tree, update, ignore = [None]) == dict(a = 1, b = np.nan, c = 0)
    >>> assert tree_update(tree, update, ignore = [None, np.nan]) == dict(a = 1, b = 2, c = 0)
    >>> assert tree_update(tree, update, ignore = [None, np.nan, 0]) == tree

    :Parameters:
    ----------------
    tree : tree
        existing tree.
    update : tree
        new information.
    types : types, optional
        see tree_items. The default is (dict, Dict, dictattr).

    :Returns:
    -------
    tree
        updated tree.

    """
    items = tree_items(update, types)
    return items_to_tree(items, tree, ignore = ignore, types = types)


from pyg_base._dictable import dictable, is_dictable
from pyg_base._as_list import as_list
from pyg_base._ulist import ulist
from pyg_base._types import is_str, is_dict, is_none
from pyg_base._logger import logger
from pyg_base._dates import dt
from pyg_base._reducer import reducer
from pyg_base._decorators import wrapper

from pyg_base._inspect import argspec_add, getargspec, argspec_defaults
from operator import mul

_data = 'data'
_expiry = 'expiry'
_output = 'output'

__all__ = ['join', 'perdictable']

def _join_dictable_with_defaults(tbl_def1, tbl_def2):
    """
    performs a join while allowing for a default
    
    :Example:
    --------------
    >>> from pyg import *
    >>> tbl_def1 = dictable(a = [1,2,3], b = [4,5,6]), dict(b=0)
    >>> tbl_def2 = dictable(a = [2,3,4], c = [5,6,7]), dict(c=1)
    >>> res = _join_dictable_with_defaults(tbl_def1, tbl_def2) 
    >>> assert res[0] == dictable(a = [2,3,4,1], b = [5,6,0,4], c = [5,6,7,1])
    >>> assert res[1] == dict(b=0,c=1)


    :Parameters:
    -------------
    tbl_def1 : pair of a table and default value
    tbl_def2 : pair of a table and a default value

    :Returns:
    -------
    a pair of table and defaults

    """
    d1, def1 = tbl_def1
    d2, def2 = tbl_def2
    def1 = def1 or {}
    def2 = def2 or {}
    
    if d1 is None:
        d = d2
    elif d2 is None:
        d = d1
    else:
        d = d1 * d2
        if len(def1):
            d += (d2 / d1)(**def1)
        if len(def2):
            d += (d1 / d2)(**def2)
    defaults = {}
    defaults.update(def1)
    defaults.update(def2)
    return (d, defaults)

def _item(d, key, on = None, renames = None):
    """
    This re

    :Parameters:
    ----------------
    d : dictable or value
        value we want to extract item from
    key : TYPE
        DESCRIPTION.
    on : TYPE, optional
        DESCRIPTION. The default is None.
    renames : TYPE, optional
        DESCRIPTION. The default is None.

    :Returns:
    -------
    
    """
    on = as_list(on)
    if is_dictable(d):
        if isinstance(renames, list) and len(renames) == 1:
            renames = renames[0]
        if is_str(renames):
            d[key] = d[renames]
        elif is_dict(renames) and key in renames:
            d[key] = d[renames[key]]
        on = d.keys() & on
        if key in d.keys():
            d = d[on + [key]]
        elif _data in d.keys() and _data not in on:
            d = d.rename(**{_data: key})[on + [key]]
        elif len(d.keys()) == len(on) + 1:
            renames = (d.keys() - on)[0]
            d = d.rename(**{renames : key})[on + [key]]
        else:
            raise KeyError('Asked for key %s in %s but key not found'%(key, d.keys()))
        if len(on) == 0 or len(d) == 0:
            return d
        u = d.listby(on)
        if len(u) < len(d):
            logger.warning('WARNING: source for %s not unique per %s'%(key, on))
        return d    
    else:
        return d
        

def join(inputs, on = None, renames = None, defaults = None):
    """
    Suppose we have a function which is defined on simple numbers

    :Example:
    --------------
    >>> from pyg import *
    >>> profit = lambda amount, price: amount  * price    

    The amounts sold are available in one table and prices in another

    :Example:
    --------------
    >>> amounts = dictable(product = ['apple', 'orange', 'pear'], amount = [1,2,3])
    >>> prices = dictable(product = ['apple', 'orange', 'pear', 'banana'], price = [4,5,6,8])
    >>> join(dict(amount = amounts, price = prices), on = 'product')(profit = profit)    
    
    >>> dictable[3 x 4]
    >>> product|amount|price|profit
    >>> apple  |1     |4    |4     
    >>> orange |2     |5    |10    
    >>> pear   |3     |6    |18    
    
    :Parameters:
    ----------------
    inputs : dict
        a dict of input parameters, some of them may be dictables.
    on : str/list of str
        when we have dictables
    renames : dict, optional
        remapping. if the datasets contain multiple columns, you can say renames = dict(price = 'price_in_dollar') to tell the algo, this is the column to use
        The default is None.
    defaults : dict, optional
        Normally, an inner join is performed. However, if there is a default value/formula for when e.g. a price is missing, use this. The default is None.

    :Returns:
    -------
    dictable
        a dictable of an inner join.

    :Example: how column mapping is done
    ------------------------------------
    >>> on = 'a'
    >>> ## if there is only one column apart from keys, then it is selected:

    >>> assert join(dict(x = dictable(a = [1,2], data = [2,3])), on = on) == dictable(a = [1,2], x = [2,3])
    >>> assert join(dict(x = dictable(a = [1,2], random_name = [2,3])), on = on) == dictable(a = [1,2], x = [2,3])

    >>> ## if there are multiple columns, if variable name is there, we use it:
    >>> assert join(dict(x = dictable(a = [1,2], z = [2,3], x = [4,5])), on) == dictable(a = [1,2], x = [4,5])

    >>> ## if there are multiple columns, and 'data' is one of the columns, we use it:
    >>> assert join(dict(x = dictable(a = [1,2], z = [2,3], data = [4,5])), on) == dictable(a = [1,2], x = [4,5])

    :Example: how column mapping is done with rename
    ------------------------------------------------
    >>> with pytest.raises(KeyError):
    >>>     join(dict(x = dictable(a = [1,2], b = [2,3], c = [4,5])), on = 'a') ## pick b or c?
    >>> assert join(dict(x = dictable(a = [1,2], b = [2,3], c = [4,5])), on = 'a', renames = dict(x = 'c')) == dictable(a = [1,2,], x = [4, 5])
    
    
    :Example: joins with partial columns in some tables
    -----------------------------------------------------    
    >>> on = ['a', 'b', 'c']
    >>> a = dictable(a = [1,2,3,4], x = [1,2,3,4]) ## only column a here
    >>> b = dictable(b = [1,2,3,4], y = [1,2,3,4]) ## only column b here
    >>> c = dictable(a = [1,2,3,4], b = [1,2,3,4], c = [1,2,3,4], z = [1,2,3,4])
    >>> j = join(dict(x = a, y = b, z = c), on = ['a', 'b', 'c'])    
    >>> assert len(j) == 4 and sorted(j.keys()) == ['a', 'b', 'c', 'x', 'y', 'z']
    
    :Example: join with defaults
    ----------------------------
    If no defaults are provided, we need all variables to be present. 
    However, if we specify defaults, we left-join on that variable and insert the default value
    
    >>> x = dictable(a = [1,2,4], x = [1,2,4])
    >>> y = dictable(a = [1,2,3], x = [5,6,7])
    >>> on = 'a'
    >>> assert join(dict(x = x, y = y), on = on) == dictable(a = [1,2,], x = [1,2], y = [5,6])
    >>> assert join(dict(x = x, y = y), on = 'a', defaults = dict(x = None)) == dictable(a = [1,2,3], x = [1,2,None], y = [5,6,7])
    >>> assert join(dict(x = x, y = y), on = 'a', defaults = dict(y = 0)) == dictable(a = [1,2,4], x = [1,2,4], y = [5,6,0])
    >>> assert join(dict(x = x, y = y), on = 'a', defaults = dict(x = None, y = 0)) == dictable(a = [1,2,3,4], x = [1,2,None,4], y = [5,6,7,0])

    """
    
    defaults = {} if defaults is None else defaults
    defaults = {k:v for k,v in defaults.items() if k in inputs}
    renames = renames or {}
    seq = {key: _item(d, key, on = on.get(key) if is_dict(on) else on, renames = renames) for key, d in inputs.items()}
    non_dictables = {k : [v] for k, v in seq.items() if not is_dictable(v)}
    dictables = {k : v for k, v in seq.items() if is_dictable(v)}
    if len(dictables) == 0:
        return dictable(non_dictables)
    no_defaults = {k:v for k,v in dictables.items() if not k in defaults}
    tbl1 = reducer(mul, no_defaults.values())
    tbl_def1 = (tbl1, {})
    with_defaults = {k:v for k,v in dictables.items() if k in defaults}
    pairs = [(value, {key : defaults[key]}) for key, value in with_defaults.items()]
    tbl_def2 = reducer(_join_dictable_with_defaults, pairs, (None, None))
    tbl_def = _join_dictable_with_defaults(tbl_def1,tbl_def2) 
    res = tbl_def[0](**non_dictables)
    return res.sort(as_list(on))

class perdictable(wrapper):
    """
    A decorator that makes a function works per dictable and not just on original value

    :Example:
    -------    
    >>> f = lambda a, b: a+b
    >>> p = perdictable(f, on = ['key'])     
    
    The new modified function p now works the same on old values:

    :Paramaters:
    ------------
    function : callable
        A function
    
    on: str/list of str
        perform join based on these keys
    
    renames: dict 
        This tells us which column to grab from which table
    
    defaults: dict
        If a default is provided for a parameter, we will perform a left join, substituting missing values with the defaults
        
    if_none: bool, list of keys
        If historic data is None while the row has expired, should we force a recalculation? if True, will be done.
        
    output_is_input: bool, list of keys
        Some functions want their own outut to be presented to them. If you see to True, if cached values exist for these columns, these are provided to the function
        
    include_inputs:
        When we return the outputs, do you want the inputs to be included as well in the dictable.
        
    col: str
        the name of the variable output.

    :Example:
    -------
    >>> f = lambda a, b: a+b
    >>> p = perdictable(f, include_inputs = True)     
    >>> assert p(a = 1, b = 2) == 3
    >>> assert p(a = dictable(a = [1,2,3]), b = 3) == dictable(a = [1,2,3], b = 3, expiry = None, data = [4,5,6])



    # some parameters are constant, some are tables...
    
    >>> assert p(a = 1, b = dictable(key = ['a','b','c'], b = [1,2,3])) == dictable(key  = ['a', 'b', 'c'], data = [2,3,4])  

    # multiple tables... some unkeyed
    
    >>> assert p(a = dictable(a = [1,2]), b = dictable(key = ['a','b','c'], b = [1,2,3])) == dictable(key  = ['a','a', 'b', 'b', 'c','c'], data = [2,3,3,4,4,5])

    # multiple tables... all keyed
    
    >>> a = dictable(key = ['x', 'y'], data = [1,2])
    >>> b = dictable(key = ['y', 'z'], data = [3,4])
    >>> assert p(a = a, b = b) == dictable(key  = ['y'], data = [5])
    
    :Example: existing data provided using data and expiry
    -------------------------------------------------------
    >>> a = dictable(key = ['x', 'y', 'z'], data = [1,2,3])
    >>> b = dictable(key = ['x', 'y', 'z'], data = [1,3,4])
    >>> data = dictable(key = ['x', 'y'], data = ['we calculated this before', 'we calculated before but hasnt expired'])
    >>> expiry = dictable(key = ['x', 'y'], data = [dt(2000,1,1), dt(3000,1,1)])
    >>> inputs = dict(a = a, b = b)

    >>> res = p(a = a, b = b, data = data, expiry = expiry)
    >>> assert res.find_data(key = 'x').data == 'we calculated this before'
    >>> assert res.find_data(key = 'y').data == 5  # although calculated before, we recalculate as its expiry is in the future
    
    """
    def __init__(self, function = None, on = None, renames = None, defaults = None, if_none = False, output_is_input = True, col = 'data', include_inputs = False):
        super(perdictable, self).__init__(function = function, on = on, renames = renames, defaults = defaults, if_none = if_none, output_is_input = output_is_input, col = col, include_inputs  = include_inputs)
    
    @property
    def output(self):
        return getattr(self.function, 'output', ['data'])

    @property
    def fullargspec(self):
        return argspec_add(getargspec(self.function), expiry = None, **{o : None for o in self.output})
                                
    def _value_output(self, expiry = None, **inputs):
        """
        this runs a normal function that is not defined to have a dict as an output

        """
        on = ulist(as_list(self.on))
        col = self.col
        cols = ulist(as_list(col))
        inputs[_expiry] = expiry
        defaults = argspec_defaults(self.function) if self.defaults is None else self.defaults
        defaults[col] = defaults.get(col, None)
        defaults[_expiry] = defaults.get(_expiry, None)
        ds = join(inputs, on = on, renames = self.renames, defaults = defaults)        
        missing_cols = cols - ds.keys()            
        provided_cols = cols - missing_cols
        
        if len(ds) == 0:
            return inputs.get(col, None)
        rows = ds if self.output_is_input is True else ds - [key for key in provided_cols if key not in as_list(self.output_is_input)]
        
        if len(ds) == 1 and len({k : v for k, v in inputs.items() if is_dictable(v)}) == 0:
            return rows[0][self.function]
        else:                
            missing_cols = cols - ds.keys()
            ### rows to be run because value is None
            if len(missing_cols): # if any of the cols is missing
                run_if_none = [True] * len(ds)
            elif self.if_none is False:
                run_if_none = [False] * len(ds)
            else:
                nones = ds[cols if self.if_none is True else ds.keys() & as_list(self.if_none)].do(is_none)
                run_if_none = [max(row.values()) for row in nones]

            ### rows to be run because has not expired
            today = dt(0)
            run_expiry = [value is None or value>=today for value in ds[_expiry]]
            
            ## default values if function is not run
            cache = ds[col] if col in ds.keys() else [None]*len(ds)            
            values = [row[self.function] if (rin or rex) else c for row, rin, rex, c in zip(rows, run_if_none, run_expiry, cache)]            
            if self.include_inputs:
                return rows(**{col : values})
            elif len(on)>0:
                return rows[on](**{col : values})                
            else:
                return dictable(**{col : values})

    def _dict_output(self, expiry = None, **inputs):
        """
        this runs a function that has function.output, meaning we want the function to have a dict as an output with its keys given by function.output

        """
        on = ulist(as_list(self.on))
        cols = ulist(self.output)
        inputs[_expiry] = expiry
        defaults = argspec_defaults(self.function) if self.defaults is None else self.defaults
        for key in cols:
            defaults[key] = defaults.get(key, None)
        defaults[_expiry] = defaults.get(_expiry, None)
        ds = join(inputs, on = on, renames = self.renames, defaults = defaults)        
        missing_cols = cols - ds.keys()            
        provided_cols = cols - missing_cols
        if len(ds) == 0:
            return {key : inputs.get(key , None) for key in cols}        ### different from single function
        rows = ds if self.output_is_input is True else ds - [key for key in provided_cols if key not in as_list(self.output_is_input)]
        
        if len(ds) == 1 and len({k : v for k, v in inputs.items() if is_dictable(v)}) == 0:
            return rows[0][self.function]
        else:                       
            ### rows to be run because value is None
            if len(missing_cols): # if any of the cols is missing
                run_if_none = [True] * len(ds)
            elif self.if_none is False:
                run_if_none = [False] * len(ds)
            else:
                nones = ds[cols if self.if_none is True else ds.keys() & as_list(self.if_none)].do(is_none)
                run_if_none = [max(row.values()) for row in nones]

            ### rows to be run because has not expired
            today = dt(0)
            run_expiry = [value is None or value>today for value in ds[_expiry]]
            
            ## default values if function is not run
            cache = ds(**{key : None for key in missing_cols})[cols]            

            values = dictable([row[self.function] if (rin or rex) else c for row, rin, rex, c in zip(rows, run_if_none, run_expiry, cache)])
            if self.include_inputs:
                return {key: rows(**{key : value}) for key, value in values.items()}                 
            elif len(on)>0:
                pks = rows[on]
                return {key: pks(**{key : value}) for key, value in values.items()} 
            else:
                return values

    def wrapped(self, expiry = None, **inputs):
        if getattr(self.function, _output, None) is None:
            return self._value_output(expiry = expiry, **inputs)
        else:
            return self._dict_output(expiry = expiry, **inputs)
            

        
    


    




import numpy as np
import pandas as pd
from functools import partial

__all__ = ['eq', 'in_']

def _eq_attrs(x, y, attrs):
    for attr in attrs:
        if hasattr(x, attr) and not eq(getattr(x, attr), getattr(y, attr)):
            return False
    return True


def eq(x, y):
    """
    A better nan-handling equality comparison. Here is the problem:
    
    >>> import numpy as np
    >>> assert not np.nan == np.nan  ## What?
    
    The nan issue extends to np.arrays...

    >>> assert list(np.array([np.nan,2]) == np.array([np.nan,2])) == [False, True]
    
    but not to lists...

    >>> assert [np.nan] == [np.nan]
    
    But wait, if the lists are derived from np.arrays, then no equality...

    >>> assert not list(np.array([np.nan])) == list(np.array([np.nan]))
    
    The other issue is inheritance:

    >>> class FunnyDict(dict):
    >>>    def __getitem__(self, key):
    >>>        return 5
    >>> assert dict(a = 1) == FunnyDict(a=1) ## equality seems to ignore any type mismatch
    >>> assert not dict(a = 1)['a'] == FunnyDict(a = 1)['a'] 
    
    There are also issues with partial

    >>> from functools import partial
    >>> f = lambda a: a + 1    
    >>> x = partial(f, a = 1)
    >>> y = partial(f, a = 1)    
    >>> assert not x == y

    >>> import pandas as pd
    >>> import pytest
    >>> from pyg import eq
    
    >>> assert eq(np.nan, np.nan) ## That's better
    >>> assert eq(x = np.array([np.nan,2]), y = np.array([np.nan,2]))    
    >>> assert eq(np.array([np.array([1,2]),2], dtype = 'object'), np.array([np.array([1,2]),2], dtype = 'object'))
    >>> assert eq(np.array([np.nan,2]),np.array([np.nan,2]))    
    >>> assert eq(dict(a = np.array([np.array([1,2]),2], dtype = 'object')) ,  dict(a = np.array([np.array([1,2]),2], dtype = 'object')))
    >>> assert eq(dict(a = np.array([np.array([1,np.nan]),np.nan])) ,  dict(a = np.array([np.array([1,np.nan]),np.nan])))
    >>> assert eq(np.array([np.array([1,2]),dict(a = np.array([np.array([1,2]),2]))]), np.array([np.array([1,2]),dict(a = np.array([np.array([1,2]),2]))]))
    
    >>> assert not eq(dict(a = 1), FunnyDict(a=1))    
    >>> assert eq(1, 1.0)
    >>> assert eq(x = pd.DataFrame([1,2]), y = pd.DataFrame([1,2]))
    >>> assert eq(pd.DataFrame([np.nan,2]), pd.DataFrame([np.nan,2]))
    >>> assert eq(pd.DataFrame([1,np.nan], columns = ['a']), pd.DataFrame([1,np.nan], columns = ['a']))
    >>> assert not eq(pd.DataFrame([1,np.nan], columns = ['a']), pd.DataFrame([1,np.nan], columns = ['b']))
    
    """
    if x is y:
        return True
    elif isinstance(x, (tuple, list)):
        return type(x) == type(y) and len(x) == len(y) and _eq_attrs(x,y,['__shape__']) and (len(x) == 0 or min([eq(i,j) for i,j in zip(x,y)]))
    elif isinstance(x, np.ndarray):
        return type(x) == type(y) and len(x) == len(y) and _eq_attrs(x,y,['__shape__']) and (0 in x.shape or np.all(veq(x,y)))
    elif isinstance(x, (pd.DataFrame, pd.Series)):
        return type(x)==type(y) and _eq_attrs(x,y, attrs = ['__shape__', 'index', 'columns']) and (0 in x.shape or np.all(veq(x,y)))
    elif isinstance(x, dict):
        if type(x) == type(y) and len(x)==len(y):
            if len(x) == 0:
                return True
            xkey, xval = zip(*sorted(x.items()))
            ykey, yval = zip(*sorted(y.items()))
            return eq(xkey, ykey) and eq(np.array(xval, dtype='object'), np.array(yval, dtype='object'))
        else:
            return False
    elif isinstance(x, float) and np.isnan(x):
        return isinstance(y, float) and np.isnan(y)    
    elif isinstance(x, partial):
        return type(x) == type(y) and x.func == y.func and eq(x.keywords, y.keywords) and eq(x.args, y.args)
    else:
        try:
            res = x == y
            return np.all(res.__array__()) if hasattr(res, '__array__') else res
        except Exception:
            return False # if you really have no == supported, the two items are not the same

veq = np.vectorize(eq)


def in_(x, seq):
    """
    Evaluates if x is in seq, avoiding issues such as these:

    >>> s = pd.Series([1,2,3])
    >>> with pytest.raises(ValueError):
    >>>     s in [None]    
    >>> assert not in_(s, [None])
    >>> assert in_(s, [None, s])    
    """
    for s in seq:
        if eq(x, s):
            return True
    return False



from typing import Awaitable
from pyg_base._decorators import wrapper
import asyncio
__all__ = ['waiter', 'async_wrapper']




# args = [waiter(f(x) if x%2 ==0 else x) for x in range(10)]
# args = [waiter(f(x)) for x in range(10)]
# res = await asyncio.gather(*args)



async def waiter(value):
    """
    serves awaited functions coroutines on a plate...

    :Example:
    ---------
    
    >>> def f(x):
    >>>     return x * 2

    >>> values = await waiter([f(i) for i in range(10)])
    >>> assert values == [0, 2, 4, 6, 8, 10, 12, 14, 16, 18]    
    >>> 
    >>> values = {i : f(i) for i in range(5)}
    >>> assert await(waiter(values)) == {0: 0, 1: 2, 2: 4, 3: 6, 4: 8}

    but also for async functions...
    
    >>> from pyg import *
    >>> import numpy as np
    >>> async def f(x):
    >>>     n = np.random.randint(5)
    >>>     await asyncio.sleep(n)
    >>>     print(x,'sleeping concurrently for ', n, 'seconds')
    >>>     return x * 3
    
    >>> values = await waiter([f(i) for i in range(10)])
    >>> assert values == [0, 3, 6, 9, 12, 15, 18, 21, 24, 27]
    >>> values = {i : f(i) for i in range(5)}
    >>> assert await(waiter(values)) == {0: 0, 1: 3, 2: 6, 3: 9, 4: 12}
    
    
    >>> values = [{j: f(i*j) for j in range(i)} for i in range(10)]
    >>> values = await waiter(values)
    >>> assert values == [{j: i*j*3 for j in range(i)} for i in range(10)]

    """
    if isinstance(value, (list,tuple)):
        values = await asyncio.gather(*[waiter(v) for v in value])
        return type(value)(values)
    elif isinstance(value, dict):
        values = await asyncio.gather(*[waiter(v) for v in value.values()])
        return type(value)(dict(zip(value.keys(), values))) 
    elif isinstance(value, Awaitable):
        return await value
    else:
        return value


class async_wrapper(wrapper):
    """
    Create a generic wrapper framework for wrapping both outright and async functions.
    
    :Example:
    ---------
    >>> class times2(async_wrapper):
    >>>     async def wrapped(self, *args, **kwargs):
    >>>         original_value = await waiter(self.function(*args, **kwargs))
    >>>         return original_value * 2
    
    >>> @times2
    >>> def f(x):
    >>>     return x+1
    >>> assert await f(2) == 6

    but the same decorator can be used for async functions too:

    >>> @times2
    >>> async def async_f(x):
    >>>     return x+1
    >>> assert await async_f(2) == 6
    
    """
    async def __call__(self, *args, **kwargs):
        if self.function is None and len(args) == 1 and len(kwargs) == 0:
            return await type(self)(function = args[0], **self._kwargs)
        else:
            return await waiter(getattr(self, 'wrapped', self.function)(*args, **kwargs))




from pyg_base._cell import cell, UPDATED, GRAPH, _updated, _data, cell_func, cell_output, cell_load, cell_go, _cell_item, cell_item
from pyg_base._inspect import getargspec, getcallargs
from pyg_base._waiter import waiter
from pyg_base._logger import logger
from pyg_base._dag import get_DAG, topological_sort
import datetime
from typing import Awaitable


__all__ = ['acell_func', 'acell', 'acell_load', 'acell_go']



async def acell_go(value, go = 0, mode = 0):
    """
    cell_go makes a cell run (using cell.go(go)) and returns the calculated cell.
    If value is not a cell, value is returned.    

    :Parameters:
    ----------------
    value : cell
        The cell (or anything else).
    go : int
        same inputs as per cell.go(go).
        0: run if cell.run() is True
        1: run this cell regardless, run parent cells only if they need to calculate too
        n: run this cell & its nth parents regardless. 
        
    :Returns:
    -------
    The calculated cell
    
    :Example: calling non-cells
    ----------------------------------------------
    >>> assert cell_go(1) == 1
    >>> assert cell_go(dict(a=1,b=2)) == dict(a=1,b=2)

    :Example: calling cells
    ------------------------------------------
    >>> c = cell(lambda a, b: a+b, a = 1, b = 2)
    >>> assert cell_go(c) == c(data = 3)

    """
    if isinstance(value, cell):
        res = value.go(go = go, mode = mode)
        if isinstance(res, Awaitable):
            return await res
        else:        
            return res
    else:
        if isinstance(value, dict):
            res = await waiter({k: acell_go(v, go = go, mode = mode) for k, v in value.items()})
            return type(value)(res) 
        elif isinstance(value, (list, tuple)):
            res = await waiter([acell_go(v, go = go, mode = mode) for v in value])
            return type(value)(res)
        else:
            return await waiter(value)

async def acell_load(value, mode = 0):
    """
    loads a cell from a database or memory and return its updated values

    :Parameters:
    ----------------
    value : cell
        The cell (or anything else).
    mode: 1/0/-1
        Used by cell.load(mode) -1 = no loading, 0 = load if available, 1 = throw an exception in unable to load

    :Returns:
    -------
    A loaded cell
    
    """
    if isinstance(value, cell):
        result = value.load(mode)
        if isinstance(result, Awaitable):
            return await result
        else:
            return result
    elif isinstance(value, (list, tuple)):
        return await waiter([acell_load(v, mode) for v in value])
    elif isinstance(value, dict):
        return await waiter({k : acell_load(v, mode) for k, v in value.items()})
    elif isinstance(value, Awaitable):
        return await value
    else:
        return value


class acell_func(cell_func):
    """
    acell_func performs the same (async) function as cell_func, loading and ensuring data is there and then presenting parameters to self.function.
    
    :Example:
    --------
    >>> from pyg import * 
    >>> a = acell(add_, a = 1, b = 2)
    >>> self = acell_func(add_)
    >>> args = ();  kwargs = dict(a = a, b = a)    
    >>> res = await self(*args, **kwargs)
    >>> assert res[0] == 6    
    """
    async def __call__(self, *args, **kwargs):
        if self.function is None and len(args) == 1 and len(kwargs) == 0:
            return await type(self)(function = args[0], **self._kwargs)
        else:
            return await waiter(getattr(self, 'wrapped', self.function)(*args, **kwargs))

    async def wrapped(self, *args, **kwargs):
        """
        we load all variables together using the waiter
        then we call all loaded cells together
        """
        go = kwargs.pop('go', 0)
        mode = kwargs.pop('mode', 0)
        function_ = cell_item(await waiter(cell_go(self.function, go)))
        callargs = getcallargs(function_, *args, **kwargs)
        spec = getargspec(function_)
        arg_names = [] if spec.args is None else spec.args
        
        c = dict(callargs)
        varargs = c.pop(spec.varargs) if spec.varargs else []
        varkw = c.pop(spec.varkw) if spec.varkw else {}
        defs = spec.defaults if spec.defaults else []
        params = dict(zip(arg_names[-len(defs):], defs))
        params.update(c)

        
        loaded_args = acell_load(varargs, mode)
        loaded_varkw = {k : v if k in self.unloaded else acell_load(v, mode) for k, v in varkw.items()}
        loaded_params = {k : v if k in self.unloaded else acell_load(v, mode) for k, v in params.items()}

        loaded_args, loaded_varkw , loaded_params = await waiter((loaded_args, loaded_varkw , loaded_params))

        called_varargs = acell_go(loaded_args , go)
        called_varkw = {k : v if k in self.uncalled else acell_go(v, go) for k, v in loaded_varkw.items()}
        called_params = {k : v if k in self.uncalled else acell_go(v, go) for k, v in loaded_params.items()}

        called_varargs, called_varkw, called_params = await waiter((called_varargs, called_varkw, called_params))

        itemized_varargs = cell_item(called_varargs, _data)
        itemized_varkw = {k : v if k in self.unitemized else _cell_item(v, self.relabels[k], True) if k in self.relabels else _cell_item(v, k) for k, v in called_varkw.items()}
        itemized_params = {k : v if k in self.unitemized else _cell_item(v, self.relabels[k], True) if k in self.relabels else _cell_item(v, k) for k, v in called_params.items()}
        
        args_ = [itemized_params[arg] for arg in arg_names if arg in params] + list(itemized_varargs)
        res = await waiter(function_(*args_, **itemized_varkw))
        called_params.update(called_varkw)
        return res, itemized_varargs, called_params


class acell(cell):
    """
    asynchronous cell. 
    
    """
    _func = acell_func
    _awaitable = True

    async def __call__(self, go = 0, mode = 0, **kwargs):
        loaded = await (self + kwargs).load(mode = mode)
        return await loaded.go(go = go, mode = mode)

    async def go(self, go = 1, mode = 0, **kwargs):
        res = await (self + kwargs)._go(go = go, mode = mode)
        address = res._address
        if address in UPDATED:
            res[_updated] = UPDATED[address] 
        else: 
            res[_updated] = datetime.datetime.now()
        return await res.save()

    async def _go(self, go = 0, mode = 0):
        address = self._address
        if callable(self.function) and (go!=0 or self.run()):
            if address:
                pairs = ', '.join([("%s = '%s'" if isinstance(value, str) else "%s = %s")%(key, value) for key, value in address])
                msg = "get_cell(%s)()"%pairs
            else:
                msg = str(address)
            logger.info(msg)
            kwargs = {arg: self[arg] for arg in self._args if arg in self}
            function = self.function if isinstance(self.function, self._func) else self._func(self.function)
            mode = 0 if mode == -1 else mode
            res, called_args, called_kwargs = await function(go = go-1 if go>0 else go, mode = mode, **kwargs)
            c = self + called_kwargs
            output = cell_output(c)
            if output is None:
                c[_data] = res
            elif len(output)>1:
                for o in output:
                    c[o] = res[o]
            else:
                c[output[0]] = res
            if address:
                UPDATED[address] = datetime.datetime.now()
                GRAPH[address] = c
            return c
        else:
            if address:
                GRAPH[address] = self
            return self
    
    async def push(self):
        """
        same as push for a normal cell but done slightly differently. We want to async as many cells as possible
        but we do need to calculate parents before children
        We therefore perform a topological sort and calculate async (in parallel) within each generation while we go up the generations in a serial fashion 
        """
        me = self._address
        res = await self.go()
        generations = topological_sort(get_DAG(), [me])['gen2node']
        for i, children in sorted(generations.items())[1:]: # we skop the first generation... we just calculated it
            GRAPH.update(await waiter({child : GRAPH[child].go() for child in children}))            
            for child in children:
                if child in UPDATED:
                    del UPDATED[child]
        if me:
            del UPDATED[me]
        return res
    
    async def load(self, mode = 0):
        return super(acell, self).load(mode = mode)
    
    def _load(self, mode = 0):
        return super(acell, self).load(mode = mode)

    async def save(self):
        return super(acell, self).save()





from operator import itemgetter as _itemgetter
import sys as _sys
from pyg_base._as_list import as_list

__all__ = ['named_dict']

def _as_str(types):
    return '{%s}'%(', '.join(["'%s' : %s"%(k, str(v).split('.')[-1].replace("'", '')) for k, v in types.items()]))

def _import_value(v):
    # if isinstance(v, (list, tuple)):
    #     return sum([_import_value(w) for w in v], [])
    if isinstance(v, dict):
        return sum([_import_value(w) for w in v.values()], [])
    else:
        if '.' in v:
            vs = v.split('.')
            return ['from %s import %s'%('.'.join(vs[:-1]), vs[-1])]
        else:
            return []
    
def _as_import(types):
    return '\n'.join(set(_import_value(types)))

# def _no_dots(v):
#     if isinstance(v, (list, tuple)):
#         return [_no_dots(i) for i in v]
#     elif isinstance(v, dict):
#         return {key : _no_dots(value) for key, value in v.items()}
#     else:
#         return v.split('.')[-1]

_class_template = '''\
{types_imports}
{casts_imports}
{dict_imports}
class {typename}({basedict}):
   """
   A {typename}({keys} with defaults {defaults})
   """
   _defaults = {defaults}
   _types = {types}
   _keys = {keys}
   _casts = {casts}
   
   def __init__(self, *args, **kwargs):
       if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], dict):
           kwargs = args[0]
           args = ()
       elif len(args)>0:
           kwargs.update(dict(zip(self._keys, args)))
           args = ()
{defaults_check}
       super({typename}, self).__init__(**kwargs)
       missing_keys = set(self._keys) - set(self.keys())
       if len(missing_keys) > 0:
           raise ValueError('must also provide %s'%missing_keys)
{casts_check}
{types_check}
   def __repr__(self):
       return '{typename}%s'%super({typename}, self).__repr__()
'''

_defaults_check = '''\
       for key in self._defaults:
           if key not in kwargs:
               kwargs[key] = self._defaults[key]
'''

_types_check = '''\
       for key, value in self._types.items():
           if not isinstance(value, type) and callable(value):
               if not value(self[key]):
                   raise ValueError('parameter %s failded %s(%s) validation'%(key, value, self[key]))
           elif not isinstance(self[key], value):
               raise TypeError('parameter %s = %s must be of type %s'%(key, self[key], value))
'''

_casts_check = '''\
       for key, value in self._casts.items():
           self[key] = value(self[key])
'''

def named_dict(name, keys, defaults = {}, types = {}, casts = {}, basedict = 'pyg_base.dictattr', debug = False):
    '''
    This forms a base for all classes. It is similar to named_tuple but:
    
    - supports additional features such as casting/type checking. 
    - support default values
    
    The resulting class is a dict so can be stored in MongoDB, sent to json or be used to construct a pd.Series automatically.
    
    :Example: Simple construction
    ------------------------------
    >>> Customer = named_dict('Customer', ['name', 'date', 'balance'])
    >>> james = Customer('james', 'today', 10)
    >>> assert james['balance'] == 10
    >>> assert james.date == 'today'

    :Example: How named_dict works with json/pandas/other named_dicts 
    -----------------------------------------------------------------
    >>> class Customer(named_dict('Customer', ['name', 'date', 'balance'])):
    >>>     def add_to_balance(self, value):
    >>>         res = self.copy()
    >>>         res.balance += value
    >>>         return res

    >>> james = Customer('james', 'date', 10)    
    >>> assert james.add_to_balance(10).balance == 20
    >>> import json
    >>> assert pd.Series(james).date == 'date'
    >>> assert dict(james) == {'name': 'james', 'date': 'date', 'balance': 10}
    >>> assert json.dumps(james) == '{"name": "james", "date": "date", "balance": 10}'

    >>> class VIP(named_dict('VIP', ['name', 'date'])):
    >>>     def some_method(self):
    >>>         return 'inheritence between classes works as long as members can share'
    
    >>> vip = VIP(james)
    >>> assert vip.name == 'james' ## members moved seemlessly
    >>> assert vip.some_method() == 'inheritence between classes works as long as members can share' 

    :Example: Adding defaults 
    -------
    >>> Customer = named_dict('Customer', ['name', 'date', 'balance'], defaults = dict(balance = 0))
    >>> james = Customer('james', 'today')
    >>> assert james['balance'] == 0

    :Example: types checking
    -------
    >>> import datetime
    >>> Customer = named_dict('Customer', ['name', 'date', 'balance'], defaults = dict(balance = 0), types = dict(date = 'datetime.datetime'))
    >>> james = Customer('james', datetime.datetime.now())
    >>> assert james['balance'] == 0

    :Example: casting
    ----------------------------
    >>> Customer = named_dict('Customer', ['name', 'date', 'balance'], defaults = dict(balance = 0), types = dict(date = 'datetime.datetime'), casts = dict(balance = 'float'))
    >>> james = Customer('james', datetime.datetime.now(), balance = '10.3')
    >>> assert james['balance'] == 10.3


    :Parameters:
    ----------------
    name : str
        name of new class.
    keys : list
        list of keys that the class must have as members.
    defaults : dict, optional
        default values for the keys. The default is {}.
    types : type or callable, optional
        A test to be applied for keys either as a callable or as a type. The default is {}.
    casts : dict, optional
        function. The default is {}.
    basedict : str, optional
        name of the dict class to inherit from. The default is 'dict'.
    debug : bool, optional
        output the construction text if set to True. The default is False.

    Raises
    ------
    ValueError
        DESCRIPTION.

    :Returns:
    -------
    result : new class that inherits from a dict

    '''
    
    
    
    keys = as_list(keys)
    typename = name
    types_check = '' if len(types) == 0 else _types_check
    casts_check = '' if len(casts) == 0 else _casts_check
    defaults_check = '' if len(defaults) == 0 else _defaults_check
    
    if len(defaults) and sorted(defaults.keys()) != sorted(keys[-len(defaults):]):
        raise ValueError('default parmamters %s must come last in %s'%(defaults, keys))
    types_imports = _as_import(types)
    casts_imports = _as_import(casts)
    dict_imports = _as_import(basedict)
    
    class_definition = _class_template.format(keys = keys, 
                                              basedict = basedict.split('.')[-1], 
                                              defaults = defaults, 
                                              typename = typename, 
                                              casts = _as_str(casts), 
                                              types = _as_str(types), 
                                              types_imports = types_imports,
                                              casts_imports = casts_imports, 
                                              dict_imports = dict_imports,
                                              casts_check = casts_check,
                                              types_check = types_check,
                                              defaults_check = defaults_check)
    if debug:
        print(class_definition)
    namespace = dict(_itemgetter = _itemgetter, _property = property, basedict = basedict)
    exec(class_definition, namespace)
    result = namespace[typename]
    result.__module__ = _sys._getframe(1).f_globals.get('__name__', '__main__')
    return result
                                              
                                              
                                              
                                              
                                              
                    




from pyg_base._types import is_iterable
from pyg_base._loop import len0

__all__ = ['zipper', 'lens']

def lens(*values):
    """
    measures (and enforces) a common length across all values

    :Parameters:
    ----------------
    *values : lists

    Raises
    ------
    ValueError
        if you have values with multi lengths.

    :Returns:
    -------
    int
        common length.
        
    :Example:
    --------------
    >>> assert lens() == 0
    >>> assert lens([1,2,3], [2,4,5]) == 3
    >>> assert lens([1,2,3], [2,4,5], [6]) == 3
    """
    if len0(values) == 0:
        return 0
    all_lens = [len0(value) for value in values]
    lens = set(all_lens) - {1}
    if len(lens)>1:
        raise ValueError('found multiple lengths %s '%lens)
    return list(lens)[0] if lens else 1

def zipper(*values):
    """
    a safer version of zip

    :Examples: zipper works with single values as well as full list:
    ---------------
    >>> assert list(zipper([1,2,3], 4)) == [(1, 4), (2, 4), (3, 4)]
    >>> assert list(zipper([1,2,3], [4,5,6])) == [(1, 4), (2, 5), (3, 6)]
    >>> assert list(zipper([1,2,3], [4,5,6], [7])) ==  [(1, 4, 7), (2, 5, 7), (3, 6, 7)]
    >>> assert list(zipper([1,2,3], [4,5,6], None)) ==  [(1, 4, None), (2, 5, None), (3, 6, None)]
    >>> assert list(zipper((1,2,3), np.array([4,5,6]), None)) ==  [(1, 4, None), (2, 5, None), (3, 6, None)]
    
    :Examples: zipper rejects multi-length lists
    ---------------
    >>> import pytest
    >>> with pytest.raises(ValueError):
    >>>     zipper([1,2,3], [4,5])
    

    :Parameters:
    ----------------
    *values : lists 
        values to be zipped

    :Returns:
    -------
    zipped values

    """
    values = [list(value) if isinstance(value, zip) else value if is_iterable(value) else [value] for value in values]
    n = lens(*values)
    values = [value * n if len(value) == 1 else value for value in values]
    return zip(*values)




from pyg_base._zip import zipper
from pyg_base._types import is_tuples, is_tuple
from pyg_base._as_list import as_list

__all__ = ['getitem', 'callitem', 'callattr']

def getitem(value, key, *default):
    """
    gets an item, like getattr
    
    :Example:
    ---------
    >>> a = dict(a = 1)
    >>> assert getitem(a, 'a') == 1
    >>> assert getitem(a, 'b', 2) == 2
    
    >>> import pytest
    >>> with pytest.raises(KeyError):
    >>>     getitem(a, 'b') 

    """    
    if len(default):
        try:
            return value[key]
        except Exception:
            return default[0]
    else:
        return value[key]

def _args(args):
    if args is None:
        args = [()]
    if is_tuple(args):
        args = [args]
    assert is_tuples(args)
    return args
    
def callitem(value, key, args = None, kwargs = None):
    """
    gets an item and calls it

    :Example:
    ---------
    >>> c = dict(function = lambda a, b: a + b)
    >>> assert callitem(c, 'function', kwargs = dict(a = 1, b = 2)) == 3
    >>> assert callitem(c, 'function', args = (1, 2)) == 3
    
    Parameters
    ----------
    value : obj
        object that contrains an item.
    key : string
        key within object.
    args : tuple, optional
        tuple of values to be fed to function. The default is None.
    kwargs : dict, optional
        kwargs to be fed to the method. The default is None.

    """
    res = value
    args = _args(args)
    kwargs = kwargs or [{}]
    for m, a, k in zipper(as_list(key), as_list(args), as_list(kwargs)):        
        method  = res[m]
        a = a or ()
        k = k or {}
        res = method(*a, **k)
    return res

def callattr(value, attr, args = None, kwargs = None):
    """
    gets the attribute(s) from a value and calls its

    :Example:
    ---------
    >>> from pyg import *
    >>> value = Dict(function = lambda a, b: a + b)
    >>> assert callattr(value, 'function', kwargs = dict(a = 1, b = 2)) == 3
    >>> assert callattr(value, attr = 'function', args = (1, 2), kwargs = None) == 3

    >>> ts = pd.Series(np.random.normal(0,1,1000))    
    >>> assert ts.std() == callattr(ts, 'std')
    >>> assert eq(ts.ewm(com = 10).mean(), callattr(ts, ['ewm','mean'], kwargs = [{'com':10}, {}]))

    >>> d = dictable(a = [1,2,3,4,1,2], b = list('abcdef'))
    >>> assert callattr(d, ['inc', 'exc'], kwargs = [dict(a = 2), dict(b = 'f')]) == d.inc(a = 2).exc(b = 'f')

    Parameters
    ----------
    value : obj
        object that contrains an item.
    attr : string(s)
        key within object.
    args : tuple, optional
        tuple of values to be fed to function. The default is None.
    kwargs : dict, optional
        kwargs to be fed to the method. The default is None.

    """
    res = value
    args = _args(args)
    kwargs = kwargs or [{}]
    for m, a, k in zipper(as_list(attr), as_list(args), as_list(kwargs)):
        method  = getattr(res, m)
        a = a or ()
        k = k or {}
        res = method(*a, **k)
    return res



from pyg_npy import mkdir, path_name
from pyg_base._types import is_series, is_df, is_int, is_date, is_bool, is_str, is_float
from pyg_base._dates import dt2str, dt
from pyg_base._logger import logger
from pyg_base._as_list import as_list
import pandas as pd
import numpy as np
import jsonpickle as jp

_series = '_is_series'

__all__ = ['pd_to_parquet', 'pd_read_parquet']



def pd_to_parquet(value, path, compression = 'GZIP'):
    """
    a small utility to save df to parquet, extending both pd.Series and non-string columns    

    :Example:
    -------
    >>> from pyg_base import *
    >>> import pandas as pd
    >>> import pytest

    >>> df = pd.DataFrame([[1,2],[3,4]], drange(-1), columns = [0, dt(0)])
    >>> s = pd.Series([1,2,3], drange(-2))

    >>> with pytest.raises(ValueError): ## must have string column names
            df.to_parquet('c:/temp/test.parquet')

    >>> with pytest.raises(AttributeError): ## pd.Series has no to_parquet
            s.to_parquet('c:/temp/test.parquet')
    
    >>> df_path = pd_to_parquet(df, 'c:/temp/df.parquet')
    >>> series_path = pd_to_parquet(s, 'c:/temp/series.parquet')

    >>> df2 = pd_read_parquet(df_path)
    >>> s2 = pd_read_parquet(series_path)

    >>> assert eq(df, df2)
    >>> assert eq(s, s2)

    """
    if is_series(value):
        mkdir(path)
        df = pd.DataFrame(value)
        df.columns = [_series]
        try:
            df.to_parquet(path, compression = compression)
        except ValueError:
            df = pd.DataFrame({jp.dumps(k) : [v] for k,v in dict(value).items()})
            df[_series] = True
            df.to_parquet(path, compression = compression)
        return path
    elif is_df(value):
        mkdir(path)
        df = value.copy()
        df.columns = [jp.dumps(col) for col in df.columns]
        df.to_parquet(path, compression  = compression)
        return path
    else:
        return value        

def pd_read_parquet(path):
    """
    a small utility to read df/series from parquet, extending both pd.Series and non-string columns 

    :Example:
    -------
    >>> from pyg import *
    >>> import pandas as pd
    >>> import pytest

    >>> df = pd.DataFrame([[1,2],[3,4]], drange(-1), columns = [0, dt(0)])
    >>> s = pd.Series([1,2,3], drange(-2))

    >>> with pytest.raises(ValueError): ## must have string column names
            df.to_parquet('c:/temp/test.parquet')

    >>> with pytest.raises(AttributeError): ## pd.Series has no to_parquet
            s.to_parquet('c:/temp/test.parquet')
    
    >>> df_path = pd_to_parquet(df, 'c:/temp/df.parquet')
    >>> series_path = pd_to_parquet(s, 'c:/temp/series.parquet')

    >>> df2 = pd_read_parquet(df_path)
    >>> s2 = pd_read_parquet(series_path)

    >>> assert eq(df, df2)
    >>> assert eq(s, s2)

    """
    path = path_name(path)
    try:
        df = pd.read_parquet(path)
    except Exception:
        logger.warning('WARN: unable to read pd.read_parquet("%s")'%path)
        return None
    if df.columns[-1] == _series:
        if len(df.columns) == 1:
            res = df[_series]
            res.name = None
            return res
        else:
            return pd.Series({jp.loads(k) : df[k].values[0] for k in df.columns[:-1]})
    else:
        df.columns = [jp.loads(col) for col in df.columns]
        return df




from pyg_base._dictable import dictable
from pyg_base._dictattr import dictattr 
from pyg_base._types import is_str
from pyg_base._tree import tree_to_table, is_tree
from pyg_base._dict import Dict, tree_items, tree_setitem, tree_keys, tree_values, items_to_tree

__all__ = ['tree_repr']

def tree_repr(value, offset = 0):
    """
    a cleaner representation of a tree
    
    :Example:
    ---------
    >>> school = dict(pupils = dict(id1 = dict(name = 'james', surname = 'maxwell', gender = 'm'),
    >>>                   id2 = dict(name = 'adam', surname = 'smith', gender = 'm'),
    >>>                   id3 = dict(name = 'michell', surname = 'obama', gender = 'f'),
    >>>                   ),
    >>>     teachers = dict(math = dict(name = 'albert', surname = 'einstein', grade = 3),
    >>>                     english = dict(name = 'william', surname = 'shakespeare', grade = 3),
    >>>                     physics = dict(name = 'richard', surname = 'feyman', grade = 4)
    >>>                     ))

    >>> print(tree_repr(school, 4))
    >>> pupils:
    >>>     id1:
    >>>         {'name': 'james', 'surname': 'maxwell', 'gender': 'm'}
    >>>     id2:
    >>>         {'name': 'adam', 'surname': 'smith', 'gender': 'm'}
    >>>     id3:
    >>>         {'name': 'michell', 'surname': 'obama', 'gender': 'f'}
    >>> teachers:
    >>>     math:
    >>>         {'name': 'albert', 'surname': 'einstein', 'grade': 3}
    >>>     english:
    >>>         {'name': 'william', 'surname': 'shakespeare', 'grade': 3}
    >>>     physics:
    >>>         {'name': 'richard', 'surname': 'feyman', 'grade': 4}

    :Parameters:
    ----------------
    value : a tree
    
    offset : int, optional
        offset from the left for printing. The default is 0.

    :Returns:
    -------
    string
        a tree-like string representation of a dict-of-dicts.

    """
    o = ' ' * offset
    res = str(value)
    if isinstance(value, dict) and not isinstance(value, dictable) and len(value) and len(res) > 80:
        if type(value) == dict:
            res = '\n'.join(['%s:\n%s'%(k, tree_repr(v, 4)) for k, v in value.items()])
        else:
            res = '\n'.join([value.__class__.__name__] + ['%s:\n%s'%(k, tree_repr(v, 4)) for k, v in value.items()])
    elif isinstance(value, list) and len(value) and len(res)>80:
        res = '\n'.join([tree_repr(v, 0) for v in value])
    return o + res.replace('\n', '\n' + o)


            
            
    